static char *SCCS_Id = "@(#)entry.c\tver 1.2 (92/06/04 11:08:41)";

/*
 * Copyright 1992 Barry Brachman
 * See included file "copyright.h" for complete copyright information.
 */

#include "local.h"

#include <ctype.h>

/*
 * These functions deal with adding/finding/deleting entries within a page
 * and handling "indirect entries" (entries that are too big to fit into a
 * single page).
 * None of these functions do I/O directly - they use the underlying
 * page handling functions.
 * Also, they do not directly modify the file header.
 * Only insert() and delete() modify the contents of a page.
 */

/*
 * Page format:
 *
 * Directory part of page is a vector of unsigned shorts.
 *
 *		16 bits       16 bits     16 bits     16 bits          16 bits
 * +------------------------------------------------------------------------+
 * | # of entries | index last | keyoffset | keylength  |     dataoffset    |
 * +------------------------------------------------------------------------+
 * |  index next  | keyoffset  | keylength | dataoffset |    index next     |
 * +------------------------------------------------------------------------+
 * | ... ->                                                                 |
 * +------------------------------------------------------------------------+
 * |                            Free Space                                  |
 * +------------------------------------------------------------------------+
 * |                                        <- ....    | data | key + flags |
 * +------------------------------------------------------------------------+
 * |               data              | key + flags |   data   | key + flags | 
 * +------------------------------------------------------------------------+
 *
 * Each directory entry consists of a tuple:
 *     (keyoffset, keylength, dataoffset, indexnext)
 * These directory entries are kept sorted by their key, in ascending keylength
 * order and lexicographically for equal keylengths.  So, the first directory
 * entry is for the "smallest" key and the last is for the "largest".
 * The indexlast is the index for the directory entry that points to the
 * innermost key/data (i.e., the one bordering on the free space).
 * The indexnext fields form a circular list and keep track of the relative
 * positions of the key/data; that is, indexlast's indexnext is the index of
 * the directory entry for the key/data closest to the end of the page.  The
 * indexnext of that directory entry is for the key/data second from the end
 * of the page, and so on.
 *
 * Flags is one byte that indicates the alignment requirements for
 * the data and whether the data is stored in the current page or indirectly.
 * The alignment is specified at the time the data is stored.
 * If the data is stored indirectly, the data field in the current page
 * indicates the physical page number where the real content is and the
 * real size of the content.
 * Alignment requirements for the key, as specified by the user, are stored
 * but not enforced by the system.
 *
 * Here's the flag byte:
 *
 *    7   6    5    4    3    2    1    0
 * +----+----+----+----+----+----+----+----+
 * |    |    |    |    |    |    |    |    |
 * +----+----+----+----+----+----+----+----+
 *
 * Bits 0-1: data alignment - index into aligntab (set by user)
 * Bits 2-4: number of alignment bytes used by entry, stored so that the
 *           original size can be calculated (set by system)
 * Bits 5-6: key alignment - index into aligntab (set by user)
 * Bit 7:    1 -> indirect (set by system)
 */

#define DIRHEADERITEMS			2
#define DIPE					4	/* Directory Items (indexes) Per Entry */
#define FLAGLENGTH				1	/* in bytes */
#define NENTRIES(page)			(*(u_short *) page)
#define LASTLOC(page)			(*((u_short *) page + 1))
#define DIRBASE(page)			((u_short *) page + 2)
#define KEYOFFSET(dir, loc)		((dir)[loc])
#define KEYLENGTH(dir, loc)		((dir)[(loc) + 1])
#define VALUEOFFSET(dir, loc)	((dir)[(loc) + 2])
#define NEXTLOC(dir, loc)		((dir)[(loc) + 3])
#define NEXTENTRY(loc)			((loc) + DIPE)
#define PREVENTRY(loc)			((loc) - DIPE)
#define ENTRYTOLOC(entrynum)	((entrynum) * DIPE)

#define VALUE_ALIGN_COUNT_SHIFT	2		/* shift to get alignment count */
#define KEY_ALIGN_INDEX_SHIFT	5		/* shift to get key alignment */
#define VALUE_ALIGN_COUNT_MASK	0x1c	/* mask to get alignment count */
#define KEY_ALIGN_INDEX_MASK	0x60	/* mask to get key alignment bits */
#define VALUE_ALIGN_INDEX_MASK	0x03	/* Isolate alignment index bits */
#define KEY_ALIGN_MASK			0x03	/* Isolate alignment index bits */
#define VALUE_ALIGN_INDEX(DESC)	(DESC & VALUE_ALIGN_INDEX_MASK)
#define VALUE_ALIGN_COUNT(DESC)	((DESC & VALUE_ALIGN_COUNT_MASK) \
								 >> VALUE_ALIGN_COUNT_SHIFT)
#define KEY_ALIGN_INDEX(DESC)	((DESC & KEY_ALIGN_INDEX_MASK) \
								 >> KEY_ALIGN_INDEX_SHIFT)
static int aligntab[4] = { 0x00, 0x02, 0x04, 0x08 };
static int entry_debug = 0;

static int checkpage PROTO((char *page, int pagesize));
static DbmRc delete PROTO((DbmTrans *trans, PageCache **cachep, int loc,
					int delete_indirect));
static DbmRc insert PROTO((DbmTrans *trans, PageCache **cachep, int loc,
					Datum *key, Datum *value));
static int locateDirEntry PROTO((char *page, u_short *entries,
					int nentries, Datum *key, int *condp));
static int locateEntryByAge PROTO((u_short *dir, int nentries, int lastloc,
					int nth));
static void showdata PROTO((FILE *f, int max, char *p, int len));
static void showpage PROTO((char *page, int pagesize));
static DbmRc split PROTO((DbmTrans *trans, LogPagenum pagenum,
					PageCache *cache));

/*
 * Determine which page the entry should be on, read the page, and
 * return the value.
 */
DbmRc
GetValue PARAMS((trans, key, value, getvalue),
DbmTrans *trans		_AND_
Datum *key			_AND_
Datum *value		_AND_
int getvalue)
{
	int align_count, cond, loc, entrykey, entryvalue, entrynum, nentries;
	char *page;
	u_short *dir;
	EntryDesc desc;
	LogPagenum pagenum;
	Indirect *indirect;
	PageCache *cache;
	DbmRc rc;

	if ((rc = KeyToPagenum(trans, key, &pagenum)) != DBM_OK)
		return(rc);

	if ((rc = ReadPage(trans, &cache, pagenum, MUST_EXIST)) != DBM_OK)
		return((rc == DBM_PAGE_MAPPING_ERROR) ? DBM_ENTRY_NOT_FOUND : rc);

	page = (char *) cache->page;
	nentries = NENTRIES(page);
	dir = DIRBASE(page);

	if (nentries == 0)
		return(DBM_ENTRY_NOT_FOUND);

	entrynum = locateDirEntry(page, dir, nentries, key, &cond);
	if (cond)
		return(DBM_ENTRY_NOT_FOUND);

	loc = ENTRYTOLOC(entrynum);
	entrykey = KEYOFFSET(dir, loc);
	entryvalue = VALUEOFFSET(dir, loc);
	value->dptr = page + entryvalue;
	desc = (EntryDesc) page[entrykey + key->dsize];
	align_count = VALUE_ALIGN_COUNT(desc);
	value->dsize = entrykey - entryvalue - align_count;

	if (desc & INDIRECT) {
		/* Assumes alignment has been taken care of */
		indirect = (Indirect *) value->dptr;
		value->dsize = indirect->length;
		if (getvalue) {
			if (ReadIndirect(trans, indirect->pagenum, value) != DBM_OK)
				return(DBM_ENTRY_NOT_FOUND);
		}
	}

	value->desc = VALUE_ALIGN_INDEX(desc) | (desc & INDIRECT);
	key->desc = KEY_ALIGN_INDEX(desc);
	if (!getvalue)
		value->dptr = NULL;

	return(DBM_OK);
}

/*
 * Try to add an entry to the given page.
 * If an entry is bigger than the maximum, it is stored indirectly.
 * If the entry is too big to fit on its "home" page, the home page must
 * be split and the splitmap modified accordingly.
 * Handle alignment as requested.
 * It is assumed that the problem of duplicate keys has been handled
 * elsewhere; if not, it is possible for multiple instances of the same key
 * to exist on a page.
 * Return DBM_OK if successful or some specific error code.
 *
 * Passing the mode to this routine avoids having to do two locateDirEntry()
 * calls.
 */
DbmRc
StoreEntry PARAMS((trans, key, value, mode),
DbmTrans *trans			_AND_
Datum *key				_AND_
Datum *value			_AND_
DbmStoreMode mode)
{
	int cond, entrynum, loc, nentries;
	u_short *dir;
	char *page;
	LogPagenum pagenum;
	PageCache *cache;
	DbmRc rc;

	value->desc = VALUE_ALIGN_INDEX(value->desc);

 again:

	if ((rc = KeyToPagenum(trans, key, &pagenum)) != DBM_OK)
		return(rc);

	if ((rc = ReadPage(trans, &cache, pagenum, MAY_CREATE)) != DBM_OK)
		return(rc);

	page = (char *) cache->page;
	nentries = NENTRIES(page);
	dir = DIRBASE(page);

	/*
	 * 1. Look for the key in the file.
	 * 2. If it exists, delete it or report an error, depending on MODE.
	 * 3. Add the entry.
	 */
	loc = 0;
	if (nentries) {
		entrynum = locateDirEntry(page, dir, nentries, key, &cond);
		loc = (cond > 0) ? ENTRYTOLOC(entrynum + 1) : ENTRYTOLOC(entrynum);
		if (cond == 0) {
			/* Found! */
			if (mode != DBM_REPLACE)
				return(DBM_ENTRY_EXISTS);
			/*
			 * XXX A special case that might be handled here is that if the
			 * sizes of the values and alignments are the same, a simple
			 * overlay could be done.
			 */
			if ((rc = delete(trans, &cache, loc, 1)) != DBM_OK)
				return(rc);
			page = (char *) cache->page;
			nentries = NENTRIES(page);
			dir = DIRBASE(page);

			loc = 0;
			if (nentries) {
				entrynum =
					locateDirEntry(page, dir, nentries, key, &cond);
				loc = (cond > 0)
					? ENTRYTOLOC(entrynum + 1) : ENTRYTOLOC(entrynum);
			}
		}
	}

	if ((rc = insert(trans, &cache, loc, key, value)) == DBM_PAGE_MUST_SPLIT) {
		if ((rc = split(trans, pagenum, cache)) == DBM_OK)
			goto again;
	}

	if (rc != DBM_OK)
		return(rc);
	return(WritePage(trans, cache));
}

/*
 * The given logical page must be split.
 * This involves determining, for each entry in the page, whether it needs
 * to be deleted and then added to a new page.
 * Unfortunately, splitting one page may result in a cascade of splits.
 */
static DbmRc
split PARAMS((trans, pagenum, cache),
DbmTrans *trans				_AND_
LogPagenum pagenum			_AND_
PageCache *cache)
{
	int align, cond, entrynum, first, lastloc, loc, nentries;
	int newloc, newnentries, nth;
	char *newpage, *page;
	u_short *dir, *newdir;
	Datum key, value;
	LogPagenum newpagenum, splitpagenum;
	PageCache *newcache;
	DbmRc rc;

	page = (char *) cache->page;
	nentries = NENTRIES(page);
	lastloc = LASTLOC(page);
	dir = DIRBASE(page);
	
	/*
	 * 1. For each of the inhabitants of the page and the given entry,
	 *    determine a new logical page number and delete and add as necessary.
	 * 2. Mark the page as having split.
	 *
	 * It is more efficient to delete from most recent entry to oldest
	 * because of the way data must be coalesced from newest to oldest.
	 */
	nth = 0;
	first = 1;
	splitpagenum = pagenum;
	newcache = NULL;
	while (nth < nentries) {
		loc = locateEntryByAge(dir, nentries, lastloc, nth);
		key.dptr = page + KEYOFFSET(dir, loc);
		key.dsize = KEYLENGTH(dir, loc);
		key.desc = (EntryDesc) key.dptr[key.dsize];

		if ((rc = SplitPagenum(trans, &key, &newpagenum)) != DBM_OK)
			return(rc);
		if (newpagenum == pagenum) {
			nth++;
			continue;
		}

		/* Verify. */
		if (splitpagenum != pagenum) {
			if (splitpagenum != newpagenum) {
				fprintf(stderr, "split: weird newpagenum?!\n");
				return(DBM_PAGE_SPLIT_FAILED);
			}
		}
		else
			splitpagenum = newpagenum;

		/* The entry must be moved to a different page. */
		if (newcache == NULL) {
			rc = ReadPage(trans, &newcache, newpagenum, MAY_CREATE);
			if (rc != DBM_OK)
				return(rc);
		}
		newpage = (char *) newcache->page;
		newdir = DIRBASE(newpage);
		newnentries = NENTRIES(newpage);

		if (first && newnentries) {
			fprintf(stderr, "split: new page occupied?!\n");
			return(DBM_PAGE_SPLIT_FAILED);
		}
		first = 0;

		entrynum = locateDirEntry(newpage, newdir, newnentries, &key, &cond);
		newloc = (cond > 0) ? ENTRYTOLOC(entrynum + 1) : ENTRYTOLOC(entrynum);

		value.desc = (EntryDesc) key.dptr[key.dsize];
		align = VALUE_ALIGN_COUNT(value.desc);
		value.dptr = page + VALUEOFFSET(dir, loc);
		value.dsize = KEYOFFSET(dir, loc) - VALUEOFFSET(dir, loc) - align;

		rc = insert(trans, &newcache, newloc, &key, &value);
		if (rc == DBM_PAGE_MUST_SPLIT) {
			fprintf(stderr, "split: insert() failed?!\n");
			return(DBM_PAGE_FULL);
		}
		/*
		  fprintf(stderr, "Moved '%s' to logpage %d\n", key.dptr, newpagenum);
		  */
		if ((rc = delete(trans, &cache, loc, 0)) != DBM_OK)
			return(rc);
		page = (char *) cache->page;
		nentries = NENTRIES(page);
		lastloc = LASTLOC(page);
		dir = DIRBASE(page);
	}

	/* It's possible that no split took place. */
	if (!first) {
		if ((rc = WritePage(trans, cache)) != DBM_OK)
			return(rc);
		if (newcache && (rc = WritePage(trans, newcache)) != DBM_OK)
			return(rc);
	}
	if ((rc = SetSplitmapBit(trans, pagenum)) != DBM_OK)
		return(rc);
	return(DBM_OK);
}

/*
 * Return the index of the directory info for the NTH newest entry.
 * NTH == 0 means the newest entry, NTH == nentries - 1 means the oldest.
 */
static int
locateEntryByAge PARAMS((dir, nentries, lastloc, nth),
u_short *dir		_AND_
int nentries		_AND_
int lastloc			_AND_
int nth)
{
	int i, loc;

	if (nentries <= 0 || nth < 0 || nth >= nentries) {
		/* XXX Can't happen! */
		fprintf(stderr, "locateEntryByAge: bad args?!\n");
		return(-1);
	}

	loc = NEXTLOC(dir, lastloc);
	for (i = 0; i < nentries - nth - 1; i++)
		loc = NEXTLOC(dir, loc);
	return(loc);
}

/*
 * Try to insert KEY and VALUE into the give page at the given location.
 */
static DbmRc
insert PARAMS((trans, cachep, loc, key, value),
DbmTrans *trans				_AND_
PageCache **cachep			_AND_
int loc						_AND_
Datum *key					_AND_
Datum *value)
{
	int align_count, align_index, align_mask, avail, i, freep, need;
	int addindirect, keyalign, lastloc, maxentrysize, nentries, pagesize;
	u_short *dir;
	char *page;
	Indirect indirect;
	DbmRc rc;
	DbmFile *dbf;
	PageCache *cache;

	cache = *cachep;
	dbf = DBF(trans);
	pagesize = dbf->pagesize;
	page = (char *) cache->page;
	nentries = NENTRIES(page);
	lastloc = LASTLOC(page);
	dir = DIRBASE(page);

	if (nentries) {
		avail = VALUEOFFSET(dir, lastloc) -
			((nentries * DIPE) + DIRHEADERITEMS) * sizeof(u_short);
		freep =  VALUEOFFSET(dir, lastloc) - 1;
	}
	else {
		avail = pagesize - (DIRHEADERITEMS * sizeof(u_short));
		freep = pagesize - 1;
	}

	if (value->dsize > dbf->max_direct_size) {
		need = sizeof(Indirect) + key->dsize + FLAGLENGTH;
		align_index = INDIRECT_ALIGN;
		align_mask = aligntab[align_index];
		value->desc = (EntryDesc) (INDIRECT | INDIRECT_ALIGN);
		addindirect = 1;
	}
	else {
		/*
		 * This entry may be an Indirect structure.
		 */
		need = value->dsize + key->dsize + FLAGLENGTH;
		align_index = VALUE_ALIGN_INDEX(value->desc);
		align_mask = aligntab[align_index];
		addindirect = 0;
	}

	if (align_mask)
		need += (align_count = (freep - need + 1) % align_mask)
			+ (DIPE * sizeof(u_short));
	else {
		align_count = 0;
		need += (DIPE * sizeof(u_short));
	}

	if (need > avail) {
		/* Some requests can never fit. */
		maxentrysize = pagesize - DIRHEADERITEMS * sizeof(u_short)
			- aligntab[MAXALIGN] - 1 - FLAGLENGTH;
		if (need > maxentrysize)
			return(DBM_KEY_TOO_LONG);
		/* Must split - no room in this page. */
		return(DBM_PAGE_MUST_SPLIT);
	}

	if ((rc = UpdatePage(trans, cachep)) != DBM_OK)
		return(rc);
	cache = *cachep;
	page = (char *) cache->page;
	dir = DIRBASE(page);

	/*
	 * Add the entry (or an indirect reference) to this page.
	 * First the key (with the descriptor byte), then the value or
	 * indirect reference.
	 */
	if (addindirect) {
		/* Add the indirect data. */
		if ((rc = AddIndirect(trans, value, &indirect)) != DBM_OK)
			return(rc);
		value->dptr = (char *) &indirect;
		value->dsize = sizeof(Indirect);
	}

	/*
	 * Determine where the entries should be in the vector.
	 * Bubble down any entries, if necessary.
	 */
	for (i = (nentries - 1) * DIPE; i >= 0; i -= DIPE) {
		if (NEXTLOC(dir, i) >= loc)
			NEXTLOC(dir, i) += DIPE;
		if (i >= loc) {
			/* Shift right one slot */
			KEYOFFSET(dir, NEXTENTRY(i)) = KEYOFFSET(dir, i);
			KEYLENGTH(dir, NEXTENTRY(i)) = KEYLENGTH(dir, i);
			VALUEOFFSET(dir, NEXTENTRY(i)) = VALUEOFFSET(dir, i);
			NEXTLOC(dir, NEXTENTRY(i)) = NEXTLOC(dir, i);
		}
	}
	if (nentries && lastloc >= loc)
		LASTLOC(page) += DIPE;

	keyalign = (key->desc & KEY_ALIGN_MASK) << KEY_ALIGN_INDEX_SHIFT;
	page[freep] = (EntryDesc) ((value->desc & INDIRECT)
		| (align_count << VALUE_ALIGN_COUNT_SHIFT) | align_index | keyalign);
	bcopy(key->dptr, page + freep - key->dsize, key->dsize);
	freep -= key->dsize;
	KEYOFFSET(dir, loc) = freep;
	KEYLENGTH(dir, loc) = key->dsize;
	bcopy(value->dptr, page+freep - value->dsize - align_count, value->dsize);
	VALUEOFFSET(dir, loc) = freep - value->dsize - align_count;
	if (NENTRIES(page)++) {
		/*
		 * New entry points back to index of the entry for the first data area,
		 * last entry now points to the new entry.
		 */
		NEXTLOC(dir, loc) = NEXTLOC(dir, LASTLOC(page));
		NEXTLOC(dir, LASTLOC(page)) = loc;
	}
	else
		NEXTLOC(dir, loc) = loc;
	LASTLOC(page) = loc;

	if (entry_debug && checkpage(page, pagesize) < 0)
		return(DBM_PAGE_CORRUPT);
	return(DBM_OK);
}

/*
 * Binary search
 * Search through a page for an entry with the given KEY.
 * Return the entry number for the key (or where the key should be in the list
 * if it is missing) and set COND to indicate whether an exact match was found
 * or whether the key should be before (cond < 0) or after (cond > 0) the
 * returned entry number.
 */
static int
locateDirEntry PARAMS((page, dir, nentries, key, condp),
char *page				_AND_
u_short *dir			_AND_
int nentries			_AND_
Datum *key				_AND_
int *condp)
{
    int len, low, high, mid, off, cond;

	if (nentries == 0) {
		*condp = 0;
		return(0);
	}

    low = 0;
    high = nentries - 1;
    mid = 0;

    while (low <= high) {
        mid = (low + high) >> 1;
		off = mid * DIPE;
		len = KEYLENGTH(dir, off);
        if ((cond = key->dsize - len) == 0)
			cond = memcmp(key->dptr, page + KEYOFFSET(dir, off), len);
        if (cond < 0)
            high = mid - 1;
        else if (cond > 0)
            low = mid + 1;
        else
            break;
    }
	*condp = cond;
    return(mid);
}

/*
 * If the specified entry can be found, delete it.
 * This may involve updating the offsets and shuffling (and realigning)
 * the entries and deleting indirect data.
 * If the page is left with no entries, it can be freed.
 */
DbmRc
DeleteEntry PARAMS((trans, key),
DbmTrans *trans			_AND_
Datum *key)
{
	int align_count, cond, entrykey, entrynum, entryvalue, loc, nentries;
	u_short *dir;
	char *page;
	Datum value;
	LogPagenum pagenum;
	PageCache *cache;
	DbmRc rc;

	if ((rc = KeyToPagenum(trans, key, &pagenum)) != DBM_OK)
		return(rc);

	if ((rc = ReadPage(trans, &cache, pagenum, MUST_EXIST)) != DBM_OK)
		return(rc);

	page = (char *) cache->page;
	nentries = NENTRIES(page);
	dir = DIRBASE(page);

	if (nentries == 0)
		return(DBM_ENTRY_NOT_FOUND);

	entrynum = locateDirEntry(page, dir, nentries, key, &cond);
	if (cond)
		return(DBM_ENTRY_NOT_FOUND);

	loc = ENTRYTOLOC(entrynum);
	entrykey = KEYOFFSET(dir, loc);
	entryvalue = VALUEOFFSET(dir, loc);
	value.dptr = page + entryvalue;
	value.desc = (EntryDesc) page[entrykey + key->dsize];
	align_count = VALUE_ALIGN_COUNT(value.desc);
	value.dsize = entrykey - entryvalue - align_count;

	/* N.B. The value pointers are junk after this */
	if ((rc = delete(trans, &cache, loc, 1)) != DBM_OK)
		return(rc);

	return(WritePage(trans, cache));
}

/*
 * Remove the entry with starting directory index LOC from the page.
 * This is complicated by the need to preserve alignment.
 * The indirect data is deleted iff DELETE_INDIRECT is non-zero.
 */
static DbmRc
delete PARAMS((trans, cachep, loc, delete_indirect),
DbmTrans *trans				_AND_
PageCache **cachep			_AND_
int loc						_AND_
int delete_indirect)
{
	int entrykey, entrynext, entryvalue, i, keylen, lastloc, n, nentries;
	int freep, movedlast, nexttolast, pagesize;
	char *page;
	u_short *dir;
	DbmRc rc;
	EntryDesc desc;
	PageCache *cache;

	cache = *cachep;
	if ((nentries = NENTRIES(cache->page)) == 0) {
		fprintf(stderr, "delete() called when no entries?!\n");
		return(DBM_OK);
	}

	/*
	 * Since there is only a single pointer to an indirect entry,
	 * locking the home entry serves to lock the indirect part too.
	 */
	if ((rc = UpdatePage(trans, cachep)) != DBM_OK)
		return(rc);

	cache = *cachep;
	page = (char *) cache->page;
	lastloc = LASTLOC(page);
	dir = DIRBASE(page);
	entrykey = KEYOFFSET(dir, loc);
	keylen = KEYLENGTH(dir, loc);
	entryvalue = VALUEOFFSET(dir, loc);

	if (delete_indirect) {
		Indirect *indirect;

		/* If there is indirect data, delete it. */
		desc = (EntryDesc) page[entrykey + keylen];
		if (desc & INDIRECT) {
			indirect = (Indirect *) (page + entryvalue);
			if ((rc = DeleteIndirect(trans, indirect)) != DBM_OK)
				return(rc);
		}
	}

	/*
	 * Here comes the grisly stuff.
	 * Step 1.
	 *   Bubble "up" the directory and adjust pointers
	 * Step 2.
	 *   Decrement the entry count and lastindex
	 * Step 3.
	 *   Shuffle "down" all key/data info that "follows" the deleted entry
	 *   while updating the directory
	 */

	/* Keep track of the index of the successor to the deleted entry. */
	if ((entrynext = NEXTLOC(dir, loc)) > loc)
		entrynext -= DIPE;
	n = nentries * DIPE;
	for (i = 0; i < n; i += DIPE) {
		/*
		 * If the successor is the last entry, keep track of this entry's index
		 * in case the last entry is being deleted.
		 */
		if (NEXTLOC(dir, i) == lastloc)
			nexttolast = (i > loc) ? (i - DIPE) : i;

		/*
		 * If the index of this entry's successor is to the "right" of the
		 * deleted entry, update the index.
		 * If the entry's successor is the deleted entry, change the link.
		 */
		if (NEXTLOC(dir, i) > loc)
			NEXTLOC(dir, i) -= DIPE;
		else if (NEXTLOC(dir, i) == loc)
			NEXTLOC(dir, i) = entrynext;

		if (i > loc) {
			KEYOFFSET(dir, PREVENTRY(i)) = KEYOFFSET(dir, i);
			KEYLENGTH(dir, PREVENTRY(i)) = KEYLENGTH(dir, i);
			VALUEOFFSET(dir, PREVENTRY(i)) = VALUEOFFSET(dir, i);
			NEXTLOC(dir, PREVENTRY(i)) = NEXTLOC(dir, i);
		}
	}
	movedlast = 0;
	if (loc == lastloc) {
		LASTLOC(page) = nexttolast;
		movedlast = 1;
	}
	else if (lastloc > loc)
		LASTLOC(page) -= DIPE;

	lastloc = LASTLOC(page);
	NENTRIES(page)--;

	/*
	 * Here's where we finally get to use those pointers we've worked so
	 * hard to keep correct!
	 * Adjust the flag/key/value and directory entry for each entry that
	 * is stored after (at lower addresses) the deleted entry.
	 */
	i = entrynext;
	freep = entrykey + keylen;
	while (!movedlast) {
		int align_count, align_mask, align_index, descp, keyalign;
		int shuffle_key, shuffle_keylen, shuffle_value, shuffle_valuelen;

		shuffle_key = KEYOFFSET(dir, i);
		shuffle_keylen = KEYLENGTH(dir, i);
		shuffle_value = VALUEOFFSET(dir, i);

		/* Find the true size of the value. */
		desc = (EntryDesc) page[shuffle_key + shuffle_keylen];
		shuffle_valuelen =
			shuffle_key - shuffle_value - VALUE_ALIGN_COUNT(desc);

		/* fprintf(stderr, "Shuffle: %s\n", page + shuffle_key); */
		keyalign = KEY_ALIGN_INDEX(desc);
		align_index = VALUE_ALIGN_INDEX(desc);

		freep = (descp = freep) - shuffle_keylen;
		bcopy(page + shuffle_key, page + freep, shuffle_keylen);
		KEYOFFSET(dir, i) = freep;
		freep -= shuffle_valuelen;
		if (align_mask = aligntab[align_index])
			align_count = freep % align_mask;
		else
			align_count = 0;
		page[descp] = (EntryDesc) ((desc & INDIRECT)
			| (align_count << VALUE_ALIGN_COUNT_SHIFT)
				| align_index | keyalign);
		freep -= align_count;
		bcopy(page + shuffle_value, page + freep, shuffle_valuelen);
		VALUEOFFSET(dir, i) = freep--;

		if (i == lastloc)
			movedlast = 1;
		else
			i = NEXTLOC(dir, i);
	}
	if (entry_debug && checkpage(page, trans->dbf->pagesize) < 0)
		return(DBM_PAGE_CORRUPT);
	return(DBM_OK);
}

/*
 * Compute the size of the maximum value that can be stored directly.
 */
int
MaxDirectEntrySize PARAMS((dbf),
DbmFile *dbf)
{
	return(0.75 * dbf->pagesize);
}

/*
 * Prepare for iteration through all keys in the dbm.
 * Return the first key and set state information for GetNextKey().
 */
DbmRc
GetFirstKey PARAMS((trans, key),
DbmTrans *trans			_AND_
Datum *key)
{
	int mask, nentries;
	u_short *dir;
	char *page;
	EntryDesc desc;
	LogPagenum pagenum;
	PageCache *cache;
	DbmRc rc;

	mask = 0;
	if ((rc = HashToPagenum(trans, 0, &mask, &pagenum)) != DBM_OK)
		return(rc);

	DBM(trans)->iter.current = pagenum;
	DBM(trans)->iter.hash = 0;
	DBM(trans)->iter.mask = mask;
	DBM(trans)->iter.entrynum = 0;

	if ((rc = ReadPage(trans, &cache, pagenum, MUST_EXIST)) != DBM_OK) {
		if (rc == DBM_PAGE_MAPPING_ERROR)
			return(GetNextKey(trans, key));
		return(rc);
	}

	page = (char *) cache->page;
	nentries = NENTRIES(page);
	dir = DIRBASE(page);

	if (nentries == 0)
		return(GetNextKey(trans, key));
	key->dptr = page + KEYOFFSET(dir, 0);
	key->dsize = KEYLENGTH(dir, 0);
	desc = (EntryDesc) key->dptr[key->dsize];
	key->desc = KEY_ALIGN_INDEX(desc);

	return(DBM_OK);
}

/*
 * Continue iterating through all keys in the dbm by returning the
 * next key and updating the state.
 *
 * XXX Bug: if the user deletes something between calls to this function,
 * it's possible for a key to be skipped; e.g., GetNextKey() (0),
 * delete key (0), then GetNextKey() (2!).
 *
 * XXX Unlocking pages after they've been completely examined here is a tricky
 * business since we don't know if they might have been locked for some other
 * reason - need explicit user permission or maybe a flag.
 */
DbmRc
GetNextKey PARAMS((trans, key),
DbmTrans *trans			_AND_
Datum *key)
{
	int loc, mask, nentries;
	u_short *dir;
	char *page;
	EntryDesc desc;
	Iter *iter;
	LogPagenum pagenum;
	PageCache *cache;
	DbmRc rc;

	iter = &DBM(trans)->iter;
	if ((rc = ReadPage(trans, &cache, iter->current, MUST_EXIST)) != DBM_OK) {
		if (rc == DBM_PAGE_MAPPING_ERROR) {
			iter->entrynum = -1;
			return(DBM_ENTRY_NOT_FOUND);
		}
		return(rc);
	}

	page = (char *) cache->page;
	nentries = NENTRIES(page);
	dir = DIRBASE(page);

	if ((loc = ++iter->entrynum) < nentries) {
		loc *= DIPE;
		key->dptr = page + KEYOFFSET(dir, loc);
		key->dsize = KEYLENGTH(dir, loc);
		desc = (EntryDesc) key->dptr[key->dsize];
		key->desc = KEY_ALIGN_INDEX(desc);
		return(DBM_OK);
	}

	while (1) {
		if ((iter->hash = Hashinc(iter->hash, iter->mask)) == 0) {
			iter->entrynum = -1;
			return(DBM_ENTRY_NOT_FOUND);
		}

		mask = 0;
		if ((rc = HashToPagenum(trans, iter->hash, &mask, &pagenum)) != DBM_OK)
			return(rc);

		if ((rc = ReadPage(trans, &cache, pagenum, MUST_EXIST)) != DBM_OK) {
			if (rc == DBM_PAGE_MAPPING_ERROR)
				continue;
			return(rc);
		}

		page = (char *) cache->page;
		nentries = NENTRIES(page);
		dir = DIRBASE(page);

		if (nentries)
			break;
	}

	key->dptr = page + KEYOFFSET(dir, 0);
	key->dsize = KEYLENGTH(dir, 0);
	desc = (EntryDesc) key->dptr[key->dsize];
	key->desc = KEY_ALIGN_INDEX(desc);
	iter->current = pagenum;
	iter->mask = mask;
	iter->entrynum = 0;

	return(DBM_OK);
}

DbmRc
PrintPage PARAMS((dbm, tid, pagenum),
Dbm *dbm				_AND_
Tid *tid				_AND_
LogPagenum pagenum)
{
	PageCache *cache;
	DbmRc rc;
	DbmTrans trans;

	trans.dbm = dbm;
	trans.dbf = dbm->dbmfile;
	trans.tid = tid;
	if ((rc = ReadPage(&trans, &cache, pagenum, MUST_EXIST)) != DBM_OK)
		return(rc);

	showpage((char *) cache->page, dbm->dbmfile->pagesize);
	return(DBM_OK);
}

static int
checkpage PARAMS((page, pagesize),
char *page			_AND_
int pagesize)
{
	int ageloc, align_count, avail, entrykey, entrykeylen, entryvalue;
	int entryvaluelen, freep, i, j, lastloc, loc, nentries;
	int prevkey, prevkeylen;
	u_short *dir;
	EntryDesc desc;
	Indirect *indirect;

	nentries = NENTRIES(page);
	lastloc = LASTLOC(page);
	dir = DIRBASE(page);

	if (nentries) {
		avail = VALUEOFFSET(dir, lastloc) -
			((nentries * DIPE) + DIRHEADERITEMS) * sizeof(u_short);
		freep =  VALUEOFFSET(dir, lastloc) - 1;
	}
	else {
		avail = pagesize - (DIRHEADERITEMS * sizeof(u_short));
		freep = pagesize - 1;
	}

	for (i = 0, loc = 0; i < nentries; i++, loc += DIPE) {
		entrykey = KEYOFFSET(dir, loc);
		entrykeylen = KEYLENGTH(dir, loc);
		entryvalue = VALUEOFFSET(dir, loc);
		desc = (EntryDesc) page[entrykey + entrykeylen];
		align_count = VALUE_ALIGN_COUNT(desc);
		entryvaluelen = entrykey - entryvalue - align_count;

		ageloc = NEXTLOC(dir, lastloc);
		for (j = 0; ageloc != loc; j++)
			ageloc = NEXTLOC(dir, ageloc);

		/* XXX Insert more consistency checks here */
		if (entrykey <= freep || entryvalue <= freep) {
			fprintf(stderr, "*Free space overlap*\n");
			return(-1);
		}
		if (desc & INDIRECT)
			indirect = (Indirect *) (page + entryvalue);
		if (loc > 0) {
			if (entrykeylen < prevkeylen || (entrykeylen == prevkeylen
				&& memcmp(page + prevkey, page + entrykey, entrykeylen) > 0)) {
				fprintf(stderr, "*** Directory misordered!\n");
				return(-1);
			}
		}
		prevkey = entrykey;
		prevkeylen = entrykeylen;
	}
	return(0);
}

static void
showpage PARAMS((page, pagesize),
char *page			_AND_
int pagesize)
{
	int ageloc, align_count, avail, entrykey, entrykeylen, entryvalue;
	int entryvaluelen, freep, i, j, lastloc, loc, nentries;
	u_short *dir;
	EntryDesc desc;
	Indirect *indirect;

	nentries = NENTRIES(page);
	lastloc = LASTLOC(page);
	dir = DIRBASE(page);

	if (nentries) {
		avail = VALUEOFFSET(dir, lastloc) -
			((nentries * DIPE) + DIRHEADERITEMS) * sizeof(u_short);
		freep =  VALUEOFFSET(dir, lastloc) - 1;
	}
	else {
		avail = pagesize - (DIRHEADERITEMS * sizeof(u_short));
		freep = pagesize - 1;
	}

	fprintf(stderr, "nentries = %d, last = %d, avail=%d, freep=%d\n",
			nentries, lastloc / DIPE, avail, freep);

	for (i = 0, loc = 0; i < nentries; i++, loc += DIPE) {
		entrykey = KEYOFFSET(dir, loc);
		entrykeylen = KEYLENGTH(dir, loc);
		entryvalue = VALUEOFFSET(dir, loc);
		desc = (EntryDesc) page[entrykey + entrykeylen];
		align_count = VALUE_ALIGN_COUNT(desc);
		entryvaluelen = entrykey - entryvalue - align_count;

		ageloc = NEXTLOC(dir, lastloc);
		for (j = 0; ageloc != loc; j++)
			ageloc = NEXTLOC(dir, ageloc);

		fprintf(stderr, "Direntry %d/Entry %d:\n", i, j);
		/* XXX Insert more consistency checks here */
		if (entrykey <= freep || entryvalue <= freep)
			fprintf(stderr, "*Free space overlap*\n");
		fprintf(stderr, "Keyoff=%d, Keylen=%d, Valueoff=%d, Valuelen=%d, ",
				entrykey, entrykeylen, entryvalue, entryvaluelen);
		fprintf(stderr, "Desc=%#o, Nextentry=%d\n",
				desc, NEXTLOC(dir, loc) / DIPE);
		fprintf(stderr, "Key=");
		showdata(stderr, 32, page + entrykey, entrykeylen);
		fprintf(stderr, "\n");
		if (desc & INDIRECT) {
			indirect = (Indirect *) (page + entryvalue);
			fprintf(stderr, "Indirect: page=%d, length=%d\n",
					indirect->pagenum, indirect->length);
		}
		else {
			fprintf(stderr, "Value=");
			showdata(stderr, 32, page + entryvalue, entryvaluelen);
			fprintf(stderr, "\n");
		}
		fprintf(stderr, "\n");
	}
}

static void
showdata PARAMS((f, max, p, len),
FILE *f			_AND_
int max			_AND_
char *p			_AND_
int len)
{
	int i;

	if (!max || max > len)
		max = len;
	if (f == NULL)
		f = stderr;

	fputc('\'', f);
	for (i = 0; i < max; i++) {
		if (isascii(p[i]) && isprint(p[i]) && p[i] != '\n' && p[i] != '\r')
			fputc(p[i], f);
		else
			fprintf(f, "\\%#0o", p[i] & 0377);
	}
	fputc('\'', f);
}

/*
 */
DbmRc
PrintUtilization PARAMS((dbm, tid),
Dbm *dbm		_AND_
Tid *tid)
{
	int avail, hashval, i, lastloc, loc, mask, n, nentries;
	int total_avail, total_empty;
	int total_indirect, total_indirects, total_indirect_pages;
	int total_nentries, total_overhead, total_pages, total_used;
	int entrykey, entrykeylen, entryvalue, entryvaluelen;
	int max_value_len, min_value_len, max_key_len, min_key_len;
	int total_key_len, total_value_len;
	int align_count;
	char *page;
	u_short *dir;
	DbmFile *dbf;
	DbmFileHeader *h;
	DbmRc rc;
	DbmTrans trans;
	EntryDesc desc;
	Indirect *indirect;
	LogPagenum pagenum;
	PageCache *cache, *header;

	trans.dbm = dbm;
	trans.tid = tid;
	trans.dbf = dbf = dbm->dbmfile;

	total_avail = 0;
	total_empty = 0;
	total_indirect = 0;
	total_indirects = 0;
	total_indirect_pages = 0;
	total_nentries = 0;
	total_overhead = 0;
	total_pages = 0;

	max_key_len = max_value_len = -1;
	min_key_len = min_value_len = 0xfffffff;	/* XXX Big */
	total_key_len = total_value_len = 0;

	hashval = 0;
	while (1) {
		mask = 0;
		if ((rc = HashToPagenum(&trans, hashval, &mask, &pagenum)) != DBM_OK)
			return(rc);

		rc = ReadPage(&trans, &cache, pagenum, MUST_EXIST);
		if (rc != DBM_OK && rc != DBM_PAGE_MAPPING_ERROR
			&& rc != DBM_PAGENUMBER_TOO_BIG)
			return(rc);
		if (rc != DBM_OK) {
			if ((hashval = Hashinc(hashval, mask)) == 0)
				break;
			continue;
		}

		page = (char *) cache->page;
		nentries = NENTRIES(page);
		lastloc = LASTLOC(page);
		dir = DIRBASE(page);
		if (nentries)
			avail = VALUEOFFSET(dir, lastloc) -
				((nentries * DIPE) + DIRHEADERITEMS) * sizeof(u_short);
		else
			avail = dbf->pagesize - (DIRHEADERITEMS * sizeof(u_short));

		total_pages++;
		total_avail += avail;
		if (nentries == 0)
			total_empty++;
		else {
			total_nentries += nentries;
			total_overhead += (DIRHEADERITEMS + DIPE + FLAGLENGTH) * nentries;
		}

		for (i = 0, loc = 0; i < nentries; i++, loc += DIPE) {
			entrykey = KEYOFFSET(dir, loc);
			entrykeylen = KEYLENGTH(dir, loc);
			entryvalue = VALUEOFFSET(dir, loc);
			desc = (EntryDesc) page[entrykey + entrykeylen];
			align_count = VALUE_ALIGN_COUNT(desc);

			total_overhead += align_count;

			if (desc & INDIRECT) {
				indirect = (Indirect *) (page + entryvalue);
				total_indirects++;
				entryvaluelen = indirect->length;
				total_indirect += indirect->length;
				total_indirect_pages +=
					(indirect->length + dbf->allocsize - 1)
						/ dbf->allocsize;
			}
			else
				entryvaluelen = entrykey - entryvalue - align_count;

			if (entrykeylen < min_key_len)
				min_key_len = entrykeylen;
			else if (entrykeylen > max_key_len)
				max_key_len = entrykeylen;
			total_key_len += entrykeylen;
			if (entryvaluelen < min_value_len)
				min_value_len = entryvaluelen;
			else if (entryvaluelen > max_value_len)
				max_value_len = entryvaluelen;
			total_value_len += entryvaluelen;
		}

		/* FreeLock(&trans, pagenum); */

		if ((hashval = Hashinc(hashval, mask)) == 0)
			break;
	}
	total_used = (total_pages * dbf->pagesize) - total_avail;
	printf("Logical Page Utilization:\n");
	printf("Total pages:    %8d (%d phys) \n",
		   total_pages, total_pages * dbf->allocunits);
	printf("Total entries:  %8d\n", total_nentries);
	printf("Total empty:    %8d\n", total_empty);
	printf("Total used:     %8d bytes\n", total_used);
	printf("Total free:     %8d bytes\n", total_avail);
	printf("Total overhead: %8d bytes\n", total_overhead);
	if (total_pages)
		printf("Mean bytes/page: %8.1f (%.1f%%)\n",
			   (double) total_used / (double) total_pages,
			   100.0 * total_used / (total_pages * dbf->pagesize));

	printf("\nIndirect Entries:\n");
	printf("Total entries:  %8d\n", total_indirects);
	printf("Total used:     %8d bytes\n", total_indirect);
	printf("Total pages:    %8d phys (%d bytes)\n",
		   total_indirect_pages, total_indirect_pages * dbf->allocsize);
	if (total_indirect_pages)
		printf("Mean bytes/page: %8.1f (%.1f%%)\n",
			   (double) total_indirect / (double) total_indirect_pages,
			   100.0 * total_indirect / (total_indirect_pages*dbf->allocsize));

	printf("\nHeader:\n");
	rc = TransRead(&trans, HEADER_PAGE, dbf->pagesize, &header);
	if (rc != DBM_OK)
		return(rc);
	h = (DbmFileHeader *) header->page;
	printf("Total static header: %d phys (%d bytes)\n",
		   (1 + NFREEMAP_PAGES) * dbf->allocunits,
		   (1 + NFREEMAP_PAGES) * dbf->pagesize);
	n = 0;
	for (i = 0; i < NSPLITMAP_PAGES; i++) {
		if (h->splitmap_pages[i])
			n++;
	}
	printf("Total splitmap:      %d phys (%d bytes)\n",
		   n * dbf->allocunits, n * dbf->pagesize);
	n = 0;
	for (i = 0; i < NPHYSMAP_PAGES; i++) {
		if (h->physmap_pages[i])
			n++;
	}
	printf("Total physmap:       %d phys (%d bytes)\n",
		   n * dbf->allocunits, n * dbf->pagesize);

	printf("\nItems:\n");
	printf("Key length: %8d min, %8d max, %8.1f mean\n",
			min_key_len, max_key_len,
			(double) total_key_len / (double) total_nentries);
	printf("Value length: %8d min, %8d max, %8.1f mean\n",
			min_value_len, max_value_len,
			(double) total_value_len / (double) total_nentries);

	return(DBM_OK);
}
