static char *SCCS_Id = "@(#)open.c\tver 1.3 (93/03/15 13:52:43)";

/*
 * Copyright 1992 Barry Brachman
 * See included file "copyright.h" for complete copyright information.
 */

#include "local.h"

#include <sys/file.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>

static void addToDbflist PROTO((DbmFile *dbf));
static DbmFile *allocDbf PROTO((void));
static DbmFile *findDbmFile PROTO((struct stat *statbuf));
static DbmFile *findVolatileDbmFile PROTO((char *pathname));
static void freeDbf PROTO((DbmFile *dbf));
static DbmRc initDbmFile PROTO((DbmFile *dbf, DbmConfig *config));
static int openDbmFile PROTO((DbmFile *dbf, int flags, int mode,
					   DbmConfig *config, struct stat *statbuf));
static int readDbmFileHeader PROTO((DbmFile *dbf, DbmConfig *config));
static void showDbf PROTO((DbmFile *dbf));

static DbmFile *dbflist = NULL;

DbmRc
DbmOpen PARAMS((pathname, type, config, dbm, recoveryp),
char *pathname			_AND_
DbmFileType type		_AND_
DbmConfig *config		_AND_
Dbm **dbm				_AND_
DbmRecovery **recoveryp)
{
	int flags, mode;
	char resolved_path[MAXPATHLEN];
	struct stat statbuf;
	DbmFile *dbf;
	DbmRc rc;
	static int dbm_id = 1;		/* Zero is invalid. */

	flags = O_CREAT | O_RDWR;
	mode = OPTION(config, mode, DBM_DEFAULT_MODE);

	if (type == DBM_VOLATILE) {
		if (pathname && pathname[0] != '\0'
			&& (dbf = findVolatileDbmFile(pathname)))
			dbf->open_count++;
		else {
			dbf = allocDbf();
			if (pathname == NULL || pathname[0] == '\0')
				dbf->pathname = NULL;
			else
				dbf->pathname = (char *) STRDUP(pathname);
			dbf->persistent = 0;
			dbf->st_dev = -1;
			dbf->st_ino = 0;
			dbf->fd = -1;
			if ((rc = initDbmFile(dbf, config)) != DBM_OK) {
				freeDbf(dbf);
				return(rc);
			}
			dbf->max_direct_size = MaxDirectEntrySize(dbf);

			addToDbflist(dbf);
			dbf->open_count = 1;
		}
	}
	else {
		if (stat(pathname, &statbuf) < 0
			|| (dbf = findDbmFile(&statbuf)) == NULL) {
			dbf = allocDbf();
			dbf->pathname = (char *) STRDUP(pathname);
			dbf->persistent = 1;

			if (openDbmFile(dbf, flags, mode, config, &statbuf) < 0) {
				freeDbf(dbf);
				return(DBM_OPEN_FAILED);
			}
#ifndef _IBMR2
			/* Canonicalize the pathname. */
			if (realpath(pathname, resolved_path) == NULL) {
				CLOSE(dbf->fd);
				freeDbf(dbf);
				return(DBM_OPEN_FAILED);
			}
#else
			/* For systems that don't have realpath(), pray... */
			strcpy(resolved_path, pathname);
#endif
			FREE(dbf->pathname);
			dbf->pathname = (char *) STRDUP(resolved_path);
			dbf->st_dev = statbuf.st_dev;
			dbf->st_ino = statbuf.st_ino;

			/* First open - read or create file header */
			if (readDbmFileHeader(dbf, config) < 0) {
				/* XXX */
				CLOSE(dbf->fd);
				freeDbf(dbf);
				return(DBM_OPEN_FAILED);
			}
			addToDbflist(dbf);
			dbf->open_count = 1;
		}
		else
			dbf->open_count++;
	}

	*dbm = (Dbm *) NAMEDMALLOC("Dbm", sizeof(Dbm));
	(*dbm)->dbmfile = dbf;
	(*dbm)->id = dbm_id++;
	(*dbm)->iter.current = 0;
	(*dbm)->iter.hash = 0;
	(*dbm)->iter.mask = 0;
	(*dbm)->iter.entrynum = -1;

	rc = DBM_OK;
	if (dbf->persistent && dbf->open_count == 1) {
		/* XXX This is done for new dbms... */
		rc = TransRecover(*dbm, recoveryp);
	}
	return(rc);
}

static void
addToDbflist PARAMS((dbf),
DbmFile *dbf)
{

	dbf->prev = NULL;
	if (dbf->next = dbflist)
		dbflist->prev = dbf;
	dbflist = dbf;
}

void
RemoveFromDbflist PARAMS((dbf),
DbmFile *dbf)
{

	if (dbf->next)
		(dbf->next)->prev = dbf->prev;
	if (dbf->prev)
		(dbf->prev)->next = dbf->next;
	if (dbf == dbflist)
		dbflist = dbf->next;
	freeDbf(dbf);
}

/*
 * Check to see if the persistent file has already been opened.
 */
static DbmFile *
findDbmFile PARAMS((statbuf),
struct stat *statbuf)
{
	DbmFile *dbf;

	for (dbf = dbflist; dbf != NULL; dbf = dbf->next) {
		if (dbf->st_dev == statbuf->st_dev && dbf->st_ino == statbuf->st_ino)
			return(dbf);
	}
	return(NULL);
}

/*
 * Check to see if the volatile file has already been opened.
 */
static DbmFile *
findVolatileDbmFile PARAMS((pathname),
char *pathname)
{
	DbmFile *dbf;

	for (dbf = dbflist; dbf != NULL; dbf = dbf->next) {
		if (dbf->pathname && STREQ(dbf->pathname, pathname))
			return(dbf);
	}
	return(NULL);
}

/*
 * Open a new file and try to lock it.
 */
static int
openDbmFile PARAMS((dbf, flags, mode, config, statbuf),
DbmFile *dbf			_AND_
int flags				_AND_
int mode				_AND_
DbmConfig *config		_AND_
struct stat *statbuf)
{
	struct flock flock;

	if ((dbf->fd = OPEN(dbf->pathname, flags, mode)) < 0)
		return(-1);

	/*
	 * Guard against a second instance of this code accessing the file.
	 * Re-stat the file in case it was just created.
	 * The default is to insist on locking the file.
	 *
	 * Note: locking on SunOS 4.1.2 is buggy - if a program crashes with locked
	 * files, the locks are not always released.  The lockoptional flag
	 * provides a way around this but should be used with care.
	 */
	flock.l_type = F_WRLCK;
	flock.l_whence = 0;
	flock.l_start = 0;
	flock.l_len = 0;
	if (fcntl(dbf->fd, F_SETLK, &flock) < 0) {
		if (config == NULL || !config->lockoptional) {
			CLOSE(dbf->fd);
			return(-1);
		}
#ifdef LOCK_WARN
		fprintf(stderr, "DbmOpen(): dbm '%s' is not locked!\n", dbf->pathname);
#endif
	}
	if (fstat(dbf->fd, statbuf) < 0) {
		CLOSE(dbf->fd);
		return(-1);
	}

	return(0);
}

/*
 * Load the file header.  If this is a new (empty) file, initialize
 * the file header.
 * Return 0 on success, -1 on error.
 */
static int
readDbmFileHeader PARAMS((dbf, config),
DbmFile *dbf			_AND_
DbmConfig *config)
{
	int st;
	DbmFileHeader header;
	DbmRc rc;

	/* assumes at beginning of file */
	if ((st = READ(dbf->fd, (char *) &header, sizeof(header))) == 0) {
		st = -1;
		if ((rc = initDbmFile(dbf, config)) != DBM_OK) {
			fprintf(stderr, "%s\n", DbmErrorString(rc));
			if (unlink(dbf->pathname) < 0)
				fprintf(stderr, "Couldn't unlink %s\n", dbf->pathname);
		}
		else if (SeekNRead(dbf->fd, (char *) &header, sizeof(header),
						   (off_t) 0, SEEK_SET)
				 == sizeof(header))
			st = sizeof(header);
	}

	if (st == sizeof(header)) {
		if (header.magic != MAGIC_HEADER_ID) {
			fprintf(stderr, "Magic header ID doesn't match!\n");
			return(-1);
		}
		dbf->incarnation_sec = header.incarnation_sec;
		dbf->incarnation_usec = header.incarnation_usec;
		dbf->pagesize = header.pagesize;
		dbf->allocunits = header.allocunits;
		dbf->allocsize = header.pagesize / header.allocunits;

		st = 0;
	}
	if (st == 0)
		dbf->max_direct_size = MaxDirectEntrySize(dbf);
	return(st);
}

/*
 * Empty file -- initialize
 * Note that physical page 0 is the header page and is therefore
 * unavailable for other uses.
 */
static DbmRc
initDbmFile PARAMS((dbf, config),
DbmFile *dbf			_AND_
DbmConfig *config)
{
	int i, size;
	struct timeval tp;
	DbmFileHeader *h;
	DbmRc rc;
	PageCache *headerpage;
	Tid *tid;
	DbmTrans trans;

	size = OPTION(config, pagesize, DBM_DEFAULT_PAGESIZE);
	if (size < sizeof(DbmFileHeader))
		return(DBM_PAGESIZE_TOO_SMALL);

	if ((rc = TransBegin(NULL, &tid)) != DBM_OK)
		return(rc);
	trans.dbm = NULL;
	trans.dbf = dbf;
	trans.tid = tid;
	dbf->pagesize = size;
	dbf->allocunits = OPTION(config, allocunits, DBM_DEFAULT_ALLOC_UNITS);
	dbf->allocsize = size / dbf->allocunits;

	if ((rc = TransRead(&trans, HEADER_PAGE, size, &headerpage)) != DBM_OK)
		return(rc);

	h = (DbmFileHeader *) headerpage->page;
	h->pagesize = size;
	h->allocunits = dbf->allocunits;

	h->magic = MAGIC_HEADER_ID;
	/* XXX assumed to be the size of an int! */
	gettimeofday(&tp, NULL);
	dbf->incarnation_sec = h->incarnation_sec = (int) tp.tv_sec;
	dbf->incarnation_usec = h->incarnation_usec = (int) tp.tv_usec;
	h->nfreemap_pages = 0;
	for (i = 0; i < NSPLITMAP_PAGES; i++)
		h->splitmap_pages[i] = 0;
	for (i = 0; i < NPHYSMAP_PAGES; i++)
		h->physmap_pages[i] = 0;
	headerpage->flags = PAGE_DIRTY;

	/*
	 * Need to preallocate logical page 0 as a special case.
	 */
	if ((rc = NewLogPage(&trans, 0, 1)) == DBM_OK)
		rc = TransCommit(tid);

	return(rc);
}

static DbmFile *
allocDbf NOPARAMS()
{
	int i;
	DbmFile *dbf;

	if ((dbf = (DbmFile *) NAMEDMALLOC("DbmFile", sizeof(DbmFile))) == NULL)
		return(NULL);

	dbf->pathname = NULL;
	dbf->st_dev = -1;
	dbf->st_ino = -1;
	dbf->open_count = 0;
	dbf->fd = -1;
	for (i = 0; i < LOCK_HASHTAB_SIZE; i++)
		dbf->lock_hashtab[i] = NULL;
	dbf->next = dbf->prev = NULL;
	return(dbf);
}

static void
freeDbf PARAMS((dbf),
DbmFile *dbf)
{

	if (dbf->pathname)
		FREE(dbf->pathname);
	FREE(dbf);
}

void
PrintOpen NOPARAMS()
{
	DbmFile *dbf;

	for (dbf = dbflist; dbf != NULL; dbf = dbf->next) {
		printf("pathname        = '%s'\n", dbf->pathname);
		printf("open_count      = %d\n", dbf->open_count);
		printf("fd              = %d\n", dbf->fd);
		printf("pagesize        = %d\n", dbf->pagesize);
		printf("allocunits      = %d\n", dbf->allocunits);
		printf("allocsize       = %d\n", dbf->allocsize);
		printf("max_direct_size = %d\n", dbf->max_direct_size);
	}
}

void
PrintHeaders PARAMS((tid),
Tid *tid)
{
	int i, n;
	DbmFile *dbf;
	DbmFileHeader *h;
	DbmRc rc;
	DbmTrans trans;
	PageCache *header_cache;

	for (dbf = dbflist; dbf != NULL; dbf = dbf->next) {
		trans.dbm = NULL;
		trans.dbf = dbf;
		trans.tid = tid;
		rc = TransRead(&trans, HEADER_PAGE, dbf->pagesize, &header_cache);
		if (rc != DBM_OK) {
			fprintf(stderr, "%s\n", DbmErrorString(rc));
			return;
		}
		h = (DbmFileHeader *) header_cache->page;
		printf("magic            = %d\n", h->magic);
		printf("incarnation_sec = %ld\n", h->incarnation_sec);
		printf("incarnation_usec = %d\n", h->incarnation_usec);
		printf("pagesize         = %d\n", h->pagesize);
		printf("allocunits       = %d\n", h->allocunits);
		printf("allocsize        = %d\n", h->pagesize / h->allocunits);
		printf("nfreemap_pages   = %d\n", h->nfreemap_pages);
		printf("Splitmap:\n");
		n = 0;
		for (i = 0; i < NSPLITMAP_PAGES; i++) {
			if (h->splitmap_pages[i]) {
				printf("(%2d:%6d)", i, h->splitmap_pages[i]);
				if ((++n % 7) == 0)
					printf("\n");
			}
		}
		printf("\n");
		printf("Physmap:\n");
		n = 0;
		for (i = 0; i < NPHYSMAP_PAGES; i++) {
			if (h->physmap_pages[i]) {
				printf("(%3d:%6d)", i, h->physmap_pages[i]);
				if ((++n % 6) == 0)
					printf("\n");
			}
		}
		printf("\n");
	}
}

void
ShowDbm PARAMS((dbm),
Dbm *dbm)
{
	DbmFile *dbf;

	if (dbm)
		showDbf(dbm->dbmfile);
	else {
		for (dbf = dbflist; dbf; dbf = dbf->next)
			showDbf(dbf);
	}
}

static void
showDbf PARAMS((dbf),
DbmFile *dbf)
{
	int bytes, i, longest, n, nchains, shortest, total_pages;
	DbfPageLock *pl;

	fprintf(stderr, "pathname=%s\n", dbf->pathname);
	fprintf(stderr, "persistent=%d, st_dev=%d, st_ino=%d\n",
			dbf->persistent, dbf->st_dev, dbf->st_ino);
	fprintf(stderr, "incarnation_sec = %ld, incarnation_usec = %d\n",
			dbf->incarnation_sec, dbf->incarnation_usec);
	fprintf(stderr, "open_count=%d, fd=%d\n", dbf->open_count, dbf->fd);
	fprintf(stderr, "pagesize=%d, allocunits=%d, allocsize=%d, maxdirect=%d\n",
			dbf->pagesize, dbf->allocunits, dbf->allocsize,
			dbf->max_direct_size);
	fprintf(stderr, "Pages: (hashtab size = %d)\n", LOCK_HASHTAB_SIZE);
	longest = shortest = -1;
	total_pages = 0;
	nchains = 0;
	bytes = 0;
	for (i = 0; i < LOCK_HASHTAB_SIZE; i++) {
		n = 0;
		for (pl = dbf->lock_hashtab[i]; pl; pl = pl->next) {
			if (pl->cache)
				bytes += pl->cache->nbytes;
			n++;
		}

		if (n) {
			nchains++;
			total_pages += n;
		}
		if (n > longest)
			longest = n;
		if (shortest < 0)
			shortest = n;
		else if (n < shortest)
			shortest = n;
	}
	fprintf(stderr, "bytes=%d, total_pages=%d\n", bytes, total_pages);
	fprintf(stderr, "nchains=%d, shortest=%d, longest=%d\n\n",
			nchains, shortest, longest);
}
