static char *SCCS_Id = "@(#)page.c\tver 1.2 (92/06/04 11:07:15)";

/*
 * Disk page-level I/O
 *
 * Copyright 1992 Barry Brachman
 * See included file "copyright.h" for complete copyright information.
 */

#include "local.h"

#include <netinet/in.h>			/* For byteorder stuff */

static u_int sliding_mask_tab[] = {
	0x00000000, 0x00000001, 0x00000003, 0x00000007,
	0x0000000f, 0x0000001f, 0x0000003f, 0x0000007f,
	0x000000ff, 0x000001ff, 0x000003ff, 0x000007ff,
	0x00000fff, 0x00001fff, 0x00003fff, 0x00007fff,
	0x0000ffff, 0x0001ffff, 0x0003ffff, 0x0007ffff,
	0x000fffff, 0x001fffff, 0x003fffff, 0x007fffff,
	0x00ffffff, 0x01ffffff, 0x03ffffff, 0x07ffffff,
	0x0fffffff, 0x1fffffff, 0x3fffffff, 0x7fffffff
};

static int hash PROTO((char *str, int len));
static DbmRc checkLogPagenum PROTO((DbmTrans *trans, LogPagenum logpagenum));
static DbmRc checkSplitmapBit PROTO((DbmTrans *trans, u_int bitnum));
static int log2 PROTO((int num));
static DbmRc logToPhys PROTO((DbmTrans *trans, LogPagenum pagenum,
					   PhysPagenum *physpagenum));
static DbmRc updatePhysmap PROTO((DbmTrans *trans, LogPagenum logpagenum,
						 PhysPagenum physpagenum));

/*
 * Hash the key and determine which logical page the entry must be stored in
 * if it is in the database.
 */
DbmRc
KeyToPagenum PARAMS((trans, key, logpagenum),
DbmTrans *trans				_AND_
Datum *key					_AND_
LogPagenum *logpagenum)
{
	int hashval;

	hashval = hash(key->dptr, key->dsize);
	/* fprintf(stderr, "key: '%s', hashval=%#o\n", key->dptr, hashval); */
	return(HashToPagenum(trans, hashval, NULL, logpagenum));
}

/*
 * Given a hash value, compute the logical page number for the entry by
 * examining the sequence of page splits.
 * The MASKP, if non-null, is a pointer to an integer containing a mask
 * state to use (probably from a previous call to this function).
 * This feature allows traversal of all pages.
 */
DbmRc
HashToPagenum PARAMS((trans, hashval, maskp, logpagenum),
DbmTrans *trans				_AND_
int hashval					_AND_
int *maskp					_AND_
LogPagenum *logpagenum)
{
	int i;
	u_int mask;
	LogPagenum pagenum;
	DbmRc rc;

	for (i = maskp ? *maskp : 0; i < 32; i++) {
		mask = sliding_mask_tab[i];
		pagenum = (LogPagenum) (hashval & mask);

		rc = checkSplitmapBit(trans, pagenum + mask);
		if (rc == DBM_PAGE_HASNT_SPLIT) {
			/* fprintf(stderr, "mask=%#o, pagenum=%d\n", mask, pagenum); */
			*logpagenum = pagenum;
			if (maskp)
				*maskp = i;
			return(DBM_OK);
		}
		else if (rc != DBM_PAGE_HAS_SPLIT)
			return(rc);
	}

	return(DBM_PAGE_HASH_OVERFLOW);
}

/*
 * Determine the logical page number for an entry whose page is
 * being split.
 */
DbmRc
SplitPagenum PARAMS((trans, key, logpagenum),
DbmTrans *trans				_AND_
Datum *key					_AND_
LogPagenum *logpagenum)
{
	int hashval, i;
	u_int mask;
	LogPagenum pagenum;
	DbmRc rc;

	hashval = hash(key->dptr, key->dsize);

	for (i = 0; i < 32; i++) {
		mask = sliding_mask_tab[i];
		pagenum = (LogPagenum) (hashval & mask);

		rc = checkSplitmapBit(trans, pagenum + mask);
		if (rc == DBM_PAGE_HASNT_SPLIT) {
			if (hashval & (mask + 1))
				*logpagenum = pagenum + mask + 1;
			else
				*logpagenum = pagenum;
			return(checkLogPagenum(trans, *logpagenum));
		}
		else if (rc != DBM_PAGE_HAS_SPLIT)
			return(rc);
	}

	return(DBM_PAGE_HASH_OVERFLOW);
}

static DbmRc
checkLogPagenum PARAMS((trans, logpagenum),
DbmTrans *trans			_AND_
LogPagenum logpagenum)
{
	int entries_per_page, pagesize, physmap_pagenum;

	pagesize = DBF(trans)->pagesize;
	entries_per_page = pagesize / sizeof(PhysPagenum);
	physmap_pagenum = logpagenum / entries_per_page;

	if (physmap_pagenum >= NPHYSMAP_PAGES)
		return(DBM_PAGENUMBER_TOO_BIG);

	return(DBM_OK);
}

int
Hashinc PARAMS((hashval, mask),
int hashval			_AND_
int mask)
{
    int bitnum;

    hashval &= sliding_mask_tab[mask];
    bitnum = sliding_mask_tab[mask] + 1;

    while (bitnum >>= 1) {
        if ((hashval & bitnum) == 0)
            return(hashval | bitnum);
        hashval &= ~bitnum;
	}

    return(0);
}

/*
 * Given a bit number, check if it is set in the splitmap.
 * Here's what the splitmap looks like:
 *                                0                                (bit 0)
 *                               0 1                               (bits 1-2)
 *                             0 1 2 3                             (bits 3-6)
 *                         0 1 2 3 4 5 6 7                         (bits 7-14)
 *                 0 1 2 3 4 5 6 7 8 9 A B C D E F                 (bits 15-30)
 * 0 1 2 3 4 5 6 7 8 9 A B C D E F G H I J K L M N O P Q R S T U V (bits 31-62)
 *
 * So, if bit 7 is set, then page 0 has been split 4 times and if bit 14 is
 * set, page 7 has been split once.
 */
static DbmRc
checkSplitmapBit PARAMS((trans, bitnum),
DbmTrans *trans			_AND_
u_int bitnum)
{
	int bytenum, dirbitnum, dirbytenum, dirpagenum, pagesize;
	char *p;
	PhysPagenum physpage;
	PageCache *cache, *header_cache;
	DbmFileHeader *h;
	DbmRc rc;

	pagesize = DBF(trans)->pagesize;
	bytenum = bitnum / 8;
	dirbitnum = bitnum % 8;
	dirbytenum = bytenum % pagesize;
	dirpagenum = bytenum / pagesize;

	if (dirpagenum >= NSPLITMAP_PAGES) {
		fprintf(stderr, "checkSplitmapBit: dirpagenum %d is too big!\n",
				dirpagenum);
		return(DBM_PAGENUMBER_TOO_BIG);
	}

	if ((rc = TransRead(trans, HEADER_PAGE, pagesize, &header_cache))
		!= DBM_OK)
		return(rc);
	h = (DbmFileHeader *) header_cache->page;

	if ((physpage = h->splitmap_pages[dirpagenum]) == 0) {
		/* The page has not been split yet. */
		return(DBM_PAGE_HASNT_SPLIT);
	}

	if ((rc = TransRead(trans, physpage, pagesize, &cache)) != DBM_OK)
		return(rc);
	p = (char *) cache->page;
	return((p[dirbytenum] & (1 << dirbitnum))
		   ? DBM_PAGE_HAS_SPLIT : DBM_PAGE_HASNT_SPLIT);
}

/*
 * Change the directory to note that PAGENUM has been split (again).
 */
DbmRc
SetSplitmapBit PARAMS((trans, pagenum),
DbmTrans *trans				_AND_
LogPagenum pagenum)
{
	int dirbitnum, bytenum, didalloc, dirbytenum, nbits, pagesize;
	u_int bitnum;
	char *p;
	LogPagenum dirpagenum;
	PageCache *cache, *header_cache;
	PhysPagenum physpage;
	DbmFileHeader *h;
	DbmRc rc;

	for (nbits = log2((int) pagenum << 1); nbits < 32; nbits++) {
		bitnum = pagenum + sliding_mask_tab[nbits];
		if ((rc = checkSplitmapBit(trans, bitnum)) == DBM_PAGE_HASNT_SPLIT)
			break;
		if (rc != DBM_PAGE_HAS_SPLIT)
			return(rc);
	}

	if (nbits >= 32)
		return(DBM_PAGE_HASH_OVERFLOW);

	/*
	 * Compute the bit number, byte offset, and (logical) page within the
	 * directory of the bit that represents the page in question.
	 */
	pagesize = DBF(trans)->pagesize;
	bytenum = (bitnum >> 3);
	dirbitnum = bitnum & 07;
	dirbytenum = bytenum % pagesize;
	dirpagenum = (LogPagenum) (bytenum / pagesize);

	if (dirpagenum >= NSPLITMAP_PAGES) {
		fprintf(stderr, "SetSplitmapBit: dirpagenum is too big!\n");
		return(DBM_SPLITMAP_TOO_BIG);
	}

	if ((rc = TransRead(trans, HEADER_PAGE, pagesize, &header_cache))
		!= DBM_OK)
		return(rc);
	h = (DbmFileHeader *) header_cache->page;

	didalloc = 0;
	if ((physpage = h->splitmap_pages[dirpagenum]) == 0) {
		/* The splitmap page has not been allocated yet. */
		if ((rc = AllocBytes(trans, pagesize, &physpage)) != DBM_OK)
			return(rc);
		didalloc = 1;
		if ((rc = TransUpdate(trans, &header_cache)) != DBM_OK)
			return(rc);
		h = (DbmFileHeader *) header_cache->page;
		h->splitmap_pages[dirpagenum] = physpage;
		header_cache->flags = PAGE_DIRTY;
	}

	/*
	 * If the splitmap page is not cached, read it.
	 */
	if ((rc = TransRead(trans, physpage, pagesize, &cache)) != DBM_OK)
		return(rc);

	/* XXX inefficient */
	if ((rc = TransUpdate(trans, &cache)) != DBM_OK)
		return(rc);

	p = (char *) cache->page;
	if (didalloc && (cache->flags & PAGE_NEW) == 0)
		bzero(p, pagesize);
	p[dirbytenum] |= (char) (1 << dirbitnum);
	cache->flags |= PAGE_DIRTY;
	return(DBM_OK);
}

/*
 * Read PAGENUM.
 */
DbmRc
ReadPage PARAMS((trans, cache, pagenum, mode),
DbmTrans *trans				_AND_
PageCache **cache			_AND_
LogPagenum pagenum			_AND_
ReadMode mode)
{
	int pagesize;
	DbmRc rc;
	DbmFile *dbf;
	PhysPagenum physpage;

	dbf = DBF(trans);
	pagesize = dbf->pagesize;
	if ((rc = logToPhys(trans, pagenum, &physpage)) == DBM_OK)
		return(TransRead(trans, physpage, pagesize, cache));

	if (mode != MAY_CREATE || rc != DBM_PAGE_MAPPING_ERROR)
		return(rc);

	/* Like NewLogPage(). */
	if ((rc = AllocBytes(trans, pagesize, &physpage)) != DBM_OK)
		return(rc);
	if ((rc = updatePhysmap(trans, pagenum, physpage)) != DBM_OK)
		return(rc);
	if ((rc = TransRead(trans, physpage, pagesize, cache)) != DBM_OK)
		return(rc);

	bzero((char *) (*cache)->page, pagesize);		/* XXX overkill? */
	(*cache)->flags |= PAGE_DIRTY;
	return(DBM_OK);
}

/*
 * Create an indirect entry, filling in the INDIRECT structure.
 */
DbmRc
AddIndirect PARAMS((trans, value, indirect),
DbmTrans *trans			_AND_
Datum *value			_AND_
Indirect *indirect)
{
	PhysPagenum start;
	DbmRc rc;

	if ((rc = AllocBytes(trans, value->dsize, &start)) != DBM_OK)
		return(rc);
	if ((rc = WriteIndirect(trans, start, value)) != DBM_OK)
		return(rc);

	indirect->pagenum = start;
	indirect->length = value->dsize;
	return(DBM_OK);
}

/*
 * Delete an indirect entry.
 */
DbmRc
DeleteIndirect PARAMS((trans, indirect),
DbmTrans *trans			_AND_
Indirect *indirect)
{
	DbmRc rc;

	if ((rc = FreeBytes(trans, indirect->pagenum, indirect->length)) != DBM_OK)
		return(rc);

	/*
	 * If the page has been cached, get rid of it.
	 * Since it may not be cached, ignore the return code.
	 */
	TransDelete(trans, indirect->pagenum);
	return(DBM_OK);
}

DbmRc
DeletePage PARAMS((trans, pagenum),
DbmTrans *trans			_AND_
LogPagenum pagenum)
{
	DbmRc rc;
	PhysPagenum physpage;

	if ((rc = logToPhys(trans, pagenum, &physpage)) == DBM_OK)
		rc = TransDelete(trans, physpage);
	return(rc);
}

/*
 * Read an indirect entry.
 */
DbmRc
ReadIndirect PARAMS((trans, pagenum, value),
DbmTrans *trans				_AND_
PhysPagenum pagenum			_AND_
Datum *value)
{
	DbmRc rc;
	PageCache *cache;

	if ((rc = TransRead(trans, pagenum, value->dsize, &cache)) != DBM_OK)
		return(rc);
	value->dptr = (char *) cache->page;
	return(DBM_OK);
}

/*
 * Create the cache entry for a new indirect entry.
 */
DbmRc
WriteIndirect PARAMS((trans, pagenum, value),
DbmTrans *trans				_AND_
PhysPagenum pagenum			_AND_
Datum *value)
{
	DbmRc rc;
	PageCache *cache;

	if ((rc = CacheAlloc(pagenum, value->dsize, &cache)) != DBM_OK)
		return(rc);

	if ((rc = TransUpdate(trans, &cache)) != DBM_OK)
		return(rc);
	/* XXX This copy might be avoided someday. */
	bcopy(value->dptr, (char *) cache->page, value->dsize);
	cache->flags |= PAGE_DIRTY;

	return(DBM_OK);
}

DbmRc
UpdatePage PARAMS((trans, cachep),
DbmTrans *trans			_AND_
PageCache **cachep)
{

	return(TransUpdate(trans, cachep));
}

/*
 * Write a page.
 */
DbmRc
WritePage PARAMS((trans, cache),
DbmTrans *trans			_AND_
PageCache *cache)
{

	cache->flags |= PAGE_DIRTY;
	return(TransWrite(trans, cache));
}

/*
 * Allocate new logical page PAGENUM that is NPAGES contiguous pages long
 */
DbmRc
NewLogPage PARAMS((trans, pagenum, npages),
DbmTrans *trans			_AND_
LogPagenum pagenum		_AND_
int npages)
{
	int pagesize;
	PhysPagenum physpage;
	DbmRc rc;

	if ((rc = logToPhys(trans, pagenum, &physpage)) == DBM_OK) {
		fprintf(stderr, "NewLogPage: page already maps!\n");
		return(DBM_PAGE_EXISTS);
	}
	else if (rc != DBM_PAGE_MAPPING_ERROR)
		return(rc);

	pagesize = DBF(trans)->pagesize;
	if ((rc = AllocBytes(trans, npages * pagesize, &physpage)) != DBM_OK)
		return(rc);
	if ((rc = updatePhysmap(trans, pagenum, physpage)) != DBM_OK)
		return(rc);
	return(DBM_OK);
}

/*
 * Use physmap to map a given logical page number to its physical page number.
 */
static DbmRc
logToPhys PARAMS((trans, logpagenum, physpagenum),
DbmTrans *trans					_AND_
LogPagenum logpagenum			_AND_
PhysPagenum *physpagenum)
{
	int entries_per_page, pagesize, physmap_pagenum, physmap_index;
	DbmFileHeader *h;
	DbmRc rc;
	PageCache *cache, *header_cache;
	PhysPagenum physpage, *phys_page_vec;

	/*
	 * Locate the page in the map that should contain the mapping.
	 */
	pagesize = DBF(trans)->pagesize;
	if (TransRead(trans, HEADER_PAGE, pagesize, &header_cache) != DBM_OK)
		return(DBM_PAGE_IO_ERROR);
	h = (DbmFileHeader *) header_cache->page;

	entries_per_page = pagesize / sizeof(PhysPagenum);
	physmap_pagenum = logpagenum / entries_per_page;
	physmap_index = logpagenum % entries_per_page;

	if (physmap_pagenum >= NPHYSMAP_PAGES) {
		fprintf(stderr, "logToPhys: logpagenum (%d) too big!\n", logpagenum);
		return(DBM_PAGENUMBER_TOO_BIG);
	}
	if ((physpage = h->physmap_pages[physmap_pagenum]) == 0)
		return(DBM_PAGE_MAPPING_ERROR);

	/*
	 * If the physmap page is not cached, read it.
	 */
	if ((rc = TransRead(trans, physpage, pagesize, &cache)) != DBM_OK)
		return(rc);

	/*
	 * Now just do the lookup.
	 */
	phys_page_vec = (PhysPagenum *) cache->page;
	if ((physpage = phys_page_vec[physmap_index]) == 0) {
#ifdef NOTDEF
		fprintf(stderr, "logToPhys: no entry for pagenum=%d.\n", logpagenum);
#endif NOTDEF
		return(DBM_PAGE_MAPPING_ERROR);
	}
	*physpagenum = physpage;
	return(DBM_OK);
}

/*
 * Change physmap to indicate that LOGPAGENUM maps to PHYSPAGENUM.
 * Return -1 on error, 0 otherwise.
 */
static DbmRc
updatePhysmap PARAMS((trans, logpagenum, physpagenum),
DbmTrans *trans				_AND_
LogPagenum logpagenum		_AND_
PhysPagenum physpagenum)
{
	int didalloc, entries_per_page, pagesize, physmap_pagenum, physmap_index;
	PageCache *cache, *header_cache;
	PhysPagenum physpage, *phys_page_vec;
	DbmFileHeader *h;
	DbmRc rc;

	/*
	 * Locate the page in the physmap that should contain the mapping.
	 */
	pagesize = DBF(trans)->pagesize;
	if ((rc =
		 TransRead(trans, HEADER_PAGE, pagesize, &header_cache)) != DBM_OK)
		return(rc);
	h = (DbmFileHeader *) header_cache->page;

	entries_per_page = pagesize / sizeof(PhysPagenum);
	physmap_pagenum = logpagenum / entries_per_page;
	physmap_index = logpagenum % entries_per_page;

	if (physmap_pagenum >= NPHYSMAP_PAGES) {
		fprintf(stderr, "updatePhysmap: pagenum (%d) too big!\n", logpagenum);
		return(DBM_PAGENUMBER_TOO_BIG);
	}
	didalloc = 0;
	if ((physpage = h->physmap_pages[physmap_pagenum]) == 0) {
		/* This mapping page has not yet been allocated.  Do so now. */
		if ((rc = AllocBytes(trans, pagesize, &physpage)) != DBM_OK)
			return(rc);
		if ((rc = TransUpdate(trans, &header_cache)) != DBM_OK)
			return(rc);
		didalloc = 1;
		h = (DbmFileHeader *) header_cache->page;
		h->physmap_pages[physmap_pagenum] = physpage;
		header_cache->flags |= PAGE_DIRTY;
	}

	/*
	 * If the physmap page is not cached, read it.
	 */
	if ((rc = TransRead(trans, physpage, pagesize, &cache)) != DBM_OK)
		return(rc);

	if ((rc = TransUpdate(trans, &cache)) != DBM_OK)
		return(rc);

	if (didalloc && (cache->flags & PAGE_NEW) == 0)
		bzero((char *) cache->page, pagesize);

	/*
	 * Now change the mapping.
	 */
	phys_page_vec = (PhysPagenum *) cache->page;
	phys_page_vec[physmap_index] = physpagenum;
	cache->flags |= PAGE_DIRTY;
	return(DBM_OK);
}

#ifdef EANHASH
static int
hash PARAMS((str, len),
char *str		_AND_
int len)
{
	int hashval;
	int b, hashi, i, j;
	static int hash_itbl[16] = {
		61, 57, 53, 49, 45, 41, 37, 33, 29, 25, 21, 17, 13,  9,  5,  1 };
#ifdef NOTDEF
	/* Ken's */
	{ 055,043,036,054,063,014,004,005,010,064,077,000,035,027,025,071 };
#endif /* NOTDEF */

	static u_long hash_ltbl[64] = {
		06100151277L, 06106161736L, 06452611562L, 05001724107L,
		02614772546L, 04120731531L, 04665262210L, 07347467531L,
		06735253126L, 06042345173L, 03072226605L, 01464164730L,
		03247435524L, 07652510057L, 01546775256L, 05714532133L,
		06173260402L, 07517101630L, 02431460343L, 01743245566L,
		00261675137L, 02433103631L, 03421772437L, 04447707466L,
		04435620103L, 03757017115L, 03641531772L, 06767633246L,
		02673230344L, 00260612216L, 04133454451L, 00615531516L,
		06137717526L, 02574116560L, 02304023373L, 07061702261L,
		05153031405L, 05322056705L, 07401116734L, 06552375715L,
		06165233473L, 05311063631L, 01212221723L, 01052267235L,
		06000615237L, 01075222665L, 06330216006L, 04402355630L,
		01451177262L, 02000133436L, 06025467062L, 07121076461L,
		03123433522L, 01010635225L, 01716177066L, 05161746527L,
		01736635071L, 06243505026L, 03637211610L, 01756474365L,
		04723077174L, 03642763134L, 05750130273L, 03655541561L
	};

	/*
	 * This is the EAN reimplementation of the dbm/ndbm hash function.
	 */
    hashi = 0;
    hashval = 0;

    for (i = 0; i < len; i++) {
		for (j = 0, b = str[i]; j < 8; j += 4, b >>= 4) {
			hashi += hash_itbl[ b & 0xF ];
			hashval += hash_ltbl[ hashi & 0x3F ];
		}
	}

    return(hashval);
}
#else
/*
 * From sdbm, an ndbm work-alike hashed database library
 * Author: oz@nexus.yorku.ca
 * Status: public domain.
 *
 * polynomial conversion ignoring overflows
 * [this seems to work remarkably well, in fact better
 * then the ndbm hash function. Replace at your own risk]
 * use: 65599   nice.
 *      65587   even better.
 *
 * [In one experiment, this function hashed 84165 symbols (English words
 * plus symbol table values) with no collisions. -bjb]
 */
static int
hash PARAMS((str, len),
char *str		_AND_
int len)
{
    register u_long n;

    n = 0;

#define HASHC   n = *str++ + 65587 * n

#ifdef NODUFF
    while (len--)
        n = HASHC;
#else

    if (len > 0) {
        int loop;

        loop = (len + 8 - 1) >> 3;
        switch (len & (8 - 1)) {
        case 0:
            do {
                HASHC;
            case 7: HASHC;
            case 6: HASHC;
            case 5: HASHC;
            case 4: HASHC;
            case 3: HASHC;
            case 2: HASHC;
            case 1: HASHC;
            } while (--loop);
        }
    }
#endif
	return((int) n);
}
#endif

DbmRc
PrintSplitmap PARAMS((dbm, tid, dirpagenum),
Dbm *dbm			_AND_
Tid *tid			_AND_
int dirpagenum)
{
	int b, bitnum, bytenum, n, pagesize, p2, l2;
	char *p;
	PhysPagenum physpage;
	PageCache *cache, *header_cache;
	DbmFile *dbf;
	DbmFileHeader *h;
	DbmRc rc;
	DbmTrans trans;

	if (dirpagenum >= NSPLITMAP_PAGES) {
		fprintf(stderr, "PrintSplitmap: dirpagenum is too big!\n");
		return(DBM_PAGENUMBER_TOO_BIG);
	}

	dbf = dbm->dbmfile;
	pagesize = dbf->pagesize;
	trans.dbm = dbm;
	trans.dbf = dbf;
	trans.tid = tid;
	if (TransRead(&trans, HEADER_PAGE, pagesize, &header_cache) != DBM_OK)
		return(DBM_PAGE_IO_ERROR);
	h = (DbmFileHeader *) header_cache->page;

	if ((physpage = h->splitmap_pages[dirpagenum]) == 0) {
		/* The page has not been allocated. */
		return(DBM_PAGE_HASNT_SPLIT);
	}

	if ((rc = TransRead(&trans, physpage, pagesize, &cache)) != DBM_OK)
		return(rc);
	p = (char *) cache->page;

	printf("Splitmap Physical Page %d (Page X : Y mask bits: bit Z):\n",
		   physpage);
	n = 0;
	for (bytenum = 0; bytenum < pagesize; bytenum++) {
		for (bitnum = 0; bitnum < 8; bitnum++) {
			if (p[bytenum] & (1 << bitnum)) {
				b = ((dirpagenum * pagesize) + bytenum) * 8 + bitnum + 1;
				l2 = log2(b);
				p2 = b - (1 << l2);
				printf("(%5d:%2d:%6d)", p2, l2, b - 1);
				if ((++n % 4) == 0)
					printf("\n");
			}
		}
	}
	printf("\n");
	return(DBM_OK);
}

DbmRc
PrintPhysmap PARAMS((dbm, tid, dirpagenum),
Dbm *dbm			_AND_
Tid *tid			_AND_
int dirpagenum)
{
	int entries_per_page, i, j, pagesize;
	PhysPagenum physpage, *phys_page_vec;
	PageCache *cache, *header_cache;
	DbmFileHeader *h;
	DbmRc rc;
	DbmTrans trans;

	if (dirpagenum >= NPHYSMAP_PAGES) {
		fprintf(stderr, "PrintPhysmap: dirpagenum is too big!\n");
		return(DBM_PAGENUMBER_TOO_BIG);
	}

	trans.tid = tid;
	trans.dbm = dbm;
	trans.dbf = dbm->dbmfile;
	pagesize = dbm->dbmfile->pagesize;
	if (TransRead(&trans, HEADER_PAGE, pagesize, &header_cache) != DBM_OK)
		return(DBM_PAGE_IO_ERROR);
	h = (DbmFileHeader *) header_cache->page;

	if ((physpage = h->physmap_pages[dirpagenum]) == 0) {
		/* No such map page. */
		return(DBM_PAGE_NO_PHYSMAP);
	}

	if ((rc = TransRead(&trans, physpage, pagesize, &cache)) != DBM_OK)
		return(rc);

	phys_page_vec = (PhysPagenum *) cache->page;
	entries_per_page = pagesize / sizeof(PhysPagenum);
	j = 0;

	for (i = 0; i < entries_per_page; i++) {
		if (phys_page_vec[i]) {
			printf("(%4d:%4d)",
				   dirpagenum * entries_per_page + i, phys_page_vec[i]);
			if ((++j % 7) == 0)
				printf("\n");
		}
	}
	if (j % 7)
		printf("\n");
	return(DBM_OK);
}

static int
log2 PARAMS((num),
int num)
{
	int i;

	for (i = 0; num > 1; i++)
		num >>= 1;
	return(i);
}
