/* SCCS_Id = "@(#)tdbm.h\tver 1.5 (94/06/22 10:51:56)" */

#ifndef _DBM_TDBM_H_
#define _DBM_TDBM_H_

#include <sys/types.h>

#include "ansi.h"
#include "copyright.h"

typedef u_int	PhysPagenum;
typedef u_int	LogPagenum;
typedef u_long	Hash;

/*
 * The freemap is a bitmap representation of physical page allocation.
 * Bit N is 0 if page N is free.  The LSB (bit 0) of an 8 bit byte represents
 * page N and the MSB of the same byte (bit 7) represents page (N + 7).
 * With a pagesize of 8192 bytes, one freemap page contains
 * (8192 bytes/page * 8 bits/byte) == 65536 bits/freemap, which can
 * map (65536 pages * pagesize bytes/page) == 536Mb.
 * Physical page 0 is the file header and pages INITIAL_FREEMAP_PAGE through
 * (INITIAL_FREEMAP_PAGE + NFREEMAP_PAGES - 1) are preallocated for the
 * freemap.  Pages after this last page are dynamically allocated,
 * so page (INITIAL_FREEMAP_PAGE + NFREEMAP_PAGES) will be the first physical
 * page represented in the freemap.
 * While this approach will leave several holes at the beginning of the file
 * (making it look much bigger than it really is), it simplifies the problem
 * of having the freemap allocate new space for itself.
 */
#define MAGIC_HEADER_ID			17		/* "The most random number." */
#define HEADER_PAGE				0
#define INITIAL_FREEMAP_PAGE	1
#define NFREEMAP_PAGES			10
/*
 * Since the header must fit into the first page, these values determine
 * the minimum page size.
 * For reasonably large databases, NSPLITMAP_PAGES=100 and NPHYSMAP_PAGES=400
 * are recommended, although they make the minimum page size 2048.
 */
#define NSPLITMAP_PAGES			100
#define NPHYSMAP_PAGES			1000
#define FIRST_AVAIL_PAGE		(INITIAL_FREEMAP_PAGE + NFREEMAP_PAGES)

/* These sizes must be powers of 2. */
#define LOCK_HASHTAB_SIZE			256
#define TRANSLOCK_HASHTAB_SIZE		256
#define LOCK_HASHTAB_MASK			(LOCK_HASHTAB_SIZE - 1)
#define TRANSLOCK_HASHTAB_MASK		(TRANSLOCK_HASHTAB_SIZE - 1)

typedef struct DbmFileHeader {
	int magic;
	int incarnation_sec;
	int incarnation_usec;
	/*
	 * Must be a power of 2 and should be fairly large (say, at least 2048).
	 * The size of each "page" in the file.  Should take into account
	 * best I/O block size for the filesystem (see stat()), maximum
	 * number of pages determined by the following limits, and internal
	 * fragmentation.
	 * An entry that won't fit into a page is stored outside the normal
	 * page space and suffers an extra disk read and, likely, internal
	 * fragmentation.
	 */
	int pagesize;
	/*
	 * This is the real size of each page in the file.
	 * Each bit in the freemap represents (pagesize / allocunits) bytes.
	 * For example, if pagesize is 8192 and allocunits is 8, then disk space
	 * allocation is in blocks of 1024 bytes and 8 contiguous blocks are needed
	 * to allocate one page.  As allocunits gets larger, external
	 * fragmentation tends to grow but internal fragmentation in indirectly
	 * stored data tends to decrease.
	 */
	int allocunits;
	/*
	 * The number of freemap pages allocated.
	 */
	int nfreemap_pages;
	/*
	 * The splitmap is really the .dir file of a dbm.  It is a bitmap
	 * that indicates which pages have been split (and how many times).
	 * See checkSplitmapBit() for detail.
	 * The number of bits B needed for N pages is:
	 *		B = 2 ^ ceil(ln2(N + 1)) - 1 (Chris Torek)
	 * For N = (2 ^ 16) - 1, B = (2 ^ 16) - 1 bits (== one 8Kb page)
	 * The vector is used to map a logical directory page number into
	 * a physical directory page number.
	 */
	PhysPagenum splitmap_pages[NSPLITMAP_PAGES];
	/*
	 * The physmap maps a logical page number to a starting physical page
	 * number.  The maximum number of logical pages is:
	 *             (NPHYSMAP_PAGES * pagesize) / sizeof(u_int)
	 */
	PhysPagenum physmap_pages[NPHYSMAP_PAGES];
} DbmFileHeader;

/* Flag bits for cached pages. */
#define PAGE_CLEAN		0
#define PAGE_DIRTY		1
#define PAGE_NEW		2

typedef struct PageCache {
	PhysPagenum			pagenum;
	int					nbytes;
	int					flags;
	void				*page;
	struct PageCache	*next;
} PageCache;

/*
 * Each active page of a dbm has a DbfPageLock structure.
 */
typedef struct DbfPageLock {
	PhysPagenum			pagenum;
	int					lockid;
	int					nheld;
	int					read_retained;
	int					write_retained;
	PageCache			*cache;			/* For base-level page */
	struct DbfPageLock	*prev;
	struct DbfPageLock	*next;
} DbfPageLock;

typedef DbfPageLock *LockHashTab;

/*
 * One instance of this control block exists per open dbm.
 * It only disappears on the last close.
 */
typedef struct DbmFile {
	char			*pathname;
	dev_t			st_dev;
	ino_t			st_ino;
	int				open_count;
	int				fd;

	/* Copied from header for convenience */
	int				pagesize;
	int				allocunits;
	int				allocsize;
	int				max_direct_size;
	int				incarnation_sec;
	int				incarnation_usec;

	int				persistent;		/* Volatile is a reserved word */

	LockHashTab		lock_hashtab[LOCK_HASHTAB_SIZE];
	struct DbmFile	*prev;
	struct DbmFile	*next;
} DbmFile;

/* TransLock flags */
#define CACHE_CLEAN		0
#define CACHE_SHARED	1
#define CACHE_DELETED	2

typedef enum {
	/* Consecutive values starting at zero is assumed. */
#define TRANS_RETAINED_LOCK(X)		(X > 2)
	TRANS_NOLOCK             = 0,
	TRANS_READLOCK           = 1,
	TRANS_WRITELOCK          = 2,
	TRANS_READLOCK_RETAINED  = 3,
	TRANS_WRITELOCK_RETAINED = 4
} LockType;

typedef struct TransLock {
	DbmFile				*dbf;
	DbfPageLock			*lockinfo;
	LockType			locktype;
	int					flags;
	PageCache			*cache;
	struct TransLock	*next;
} TransLock;

typedef TransLock *TransLockHashTab;

typedef enum {
	TRANS_ABORTED, TRANS_MUSTABORT, TRANS_ACTIVE
} TransState;

#define TID_FD(tid)			((tid)->active->fd)
#define TID_DBF(tid)		((tid)->active->dbf)

/* Transaction Identifier */
typedef struct Tid {
	struct ActiveTrans	*active;
	TransState			state;
	TransLockHashTab	translock_hashtab[TRANSLOCK_HASHTAB_SIZE];
	struct Tid			*parent;
	struct Tid			*children;		/* "Leftmost" child. */
	struct Tid			*prev;			/* List of siblings. */
	struct Tid			*next;
} Tid;

typedef struct TidName {
	int					hostid;
	int					start_time;	/* XXX assumes sizeof(long)==sizeof(int) */
	int					count;
} TidName;

typedef struct ActiveTrans {
	int					dbm_id;			/* A unique id. */
	DbmFile				*dbf;
	TidName				tidname;
	int					precommitted;
	int					fd;				/* For commit/recovery processing. */
	char				*pathname;		/* For commit/recovery processing. */
	Tid					*rootTid;
	struct ActiveTrans	*prev;
	struct ActiveTrans	*next;
} ActiveTrans;

/* State info for iterating through a dbm. */
typedef struct iter {
	LogPagenum	current;
	int			hash;
	int			mask;
	int			entrynum;
} Iter;

/*
 * An instance of this control block is returned when a dbm is opened.
 * A single dbm may be opened multiple times.
 */
typedef struct {
	int			id;
	DbmFile		*dbmfile;
	Iter		iter;
} Dbm;

/*
 * Optional system configuration parameters passed to DbmOpen().
 */
typedef enum { DBM_PERSISTENT, DBM_VOLATILE } DbmFileType;

typedef struct DbmConfig {
	int		mode;
	int		pagesize;
	int		allocunits;
	int		lockoptional;
} DbmConfig;

typedef struct DbmRecovery {
	char 				*pathname;
	TidName				tidname;
	struct DbmRecovery	*next;
} DbmRecovery;

/*
 * Alignments are indexes.
 */
#define ALIGN0		0			/* No alignment required */
#define ALIGN2		1			/* Align on any even address */
#define ALIGN4		2			/* Align on address divisible by 4 */
#define ALIGN8		3			/* Align on address divisible by 8 */
#define MALLOCED	4			/* Entry has been (will be) malloced */
#define MAXALIGN	ALIGN8		/* Biggest possible alignment */
#define INDIRECT	0x80		/* Entry is stored indirectly */

typedef u_char EntryDesc;

typedef struct {
    char		*dptr;
    int			dsize;
	EntryDesc	desc;
} Datum;

typedef enum { DBM_REPLACE, DBM_INSERT } DbmStoreMode;

typedef enum { DBM_ABORT, DBM_COMMIT } DbmTerminateMode;

typedef enum {
	DBM_NOERROR   = 0x0,		/* No error */
	DBM_WARNING   = 0x1,		/* Harmless or recoverable */
	DBM_INTERNAL  = 0x2,		/* Purely internal, user invisible */
	DBM_FATAL     = 0x4,		/* Cannot continue, manual intervention */
	DBM_BADNEWS   = 0x6,		/* A fatal internal error (bug) */
	DBM_CORRUPTED = 0xc			/* Fatal error, dbm may be corrupted */
} DbmErrorClass;

#define DBM_FATAL_ERROR(t)	(t & DBM_FATAL)

typedef enum {
	DBM_OK,
	DBM_BAD_ALIGNMENT,
	DBM_BAD_PATHNAME,
	DBM_CACHE_ERROR,
	DBM_CLOSE_FAILED,
	DBM_ENTRY_EXISTS,
	DBM_ENTRY_NOT_FOUND,
	DBM_FLOCK_FAILED,
	DBM_FSYNC_FAILED,
	DBM_INVALID_ARGUMENT,
	DBM_ITER_NOT_INITIALIZED,
	DBM_KEY_TOO_LONG,
	DBM_MALLOC_FAILED,
	DBM_NO_FREEMAP_SPACE,
	DBM_NO_TRANSACTION,
	DBM_NULL_KEY,
	DBM_OPEN_FAILED,
	DBM_PAGE_ALLOC_FAILED,
	DBM_PAGE_EXISTS,
	DBM_PAGE_FREE_FAILED,
	DBM_PAGE_FULL,
	DBM_PAGE_HAS_SPLIT,
	DBM_PAGE_HASH_OVERFLOW,
	DBM_PAGE_HASNT_SPLIT,
	DBM_PAGE_IO_ERROR,
	DBM_PAGE_MAPPING_ERROR,
	DBM_PAGE_MUST_SPLIT,
	DBM_PAGE_NO_PHYSMAP,
	DBM_PAGE_NOT_FOUND,
	DBM_PAGE_SPLIT_FAILED,
	DBM_PAGE_CORRUPT,
	DBM_PAGENUMBER_TOO_BIG,
	DBM_PAGESIZE_TOO_SMALL,
	DBM_RECOVER_CORRUPT,
	DBM_RECOVER_FAILED,
	DBM_RECOVER_IO_ERROR,
	DBM_SPLITMAP_TOO_BIG,
	DBM_TRANS_ABORTED,
	DBM_TRANS_IN_PROGRESS,
	DBM_TRANS_IO_ERROR,
	DBM_TRANS_LOCK_REQUEST_FAILED,
	DBM_TRANS_LOCK_PROTOCOL_ERROR,
	DBM_TRANS_NOT_TOPLEVEL,
	DBM_TRANS_PRECOMMITTED,
	DBM_UNIMPLEMENTED,
	DBM_UNLINK_FAILED,
	DBM_BAD_RECOVER_FILE,
	DBM_RECOVER_NO_DISK_SPACE,
	DBM_UNIX_ERROR
} DbmRc;

extern DbmRc DbmAbort PROTO((Tid *tid));
extern DbmRc DbmBegin PROTO((Tid *parent, Tid **child));
extern DbmRc DbmClose PROTO((Dbm *dbm));
extern DbmRc DbmCommit PROTO((Tid *tid));
extern DbmRc DbmDelete PROTO((Dbm *dbm, Tid *tid, Datum key));
extern char *DbmErrorString PROTO((DbmRc rc));
extern DbmErrorClass DbmRcClass PROTO((DbmRc rc));
extern DbmRc DbmFetch PROTO((Dbm *dbm, Tid *tid, Datum key, Datum *value));
extern DbmRc DbmFind PROTO((Dbm *dbm, Tid *tid, Datum key, Datum *value));
extern DbmRc DbmFirstkey PROTO((Dbm *dbm, Tid *tid, Datum *key));
extern DbmRc DbmNextkey PROTO((Dbm *dbm, Tid *tid, Datum *key));
extern DbmRc DbmOpen PROTO((char *pathname, DbmFileType type,
							DbmConfig *config, Dbm **dbm,
							DbmRecovery **recoveryp));
extern DbmRc DbmPrecommit PROTO((Tid *tid, TidName *tidname));
extern DbmRc DbmRecover PROTO((Dbm *dbm, DbmRecovery **recovery, Tid **tid));
extern DbmRc DbmStore PROTO((Dbm *dbm, Tid *tid, Datum key, Datum value,
							 DbmStoreMode mode));
extern DbmRc DbmTidName PROTO((Tid *tid, TidName *tidname));
#endif
