static char *SCCS_Id = "@(#)trans.c\tver 1.7 (94/06/22 10:51:52)";

/*
 * Transaction Management
 *
 * ACID - All-or-None, Consistency, Isolation, Durability
 *
 * All-or-None (Recovery):
 * Recovery is currently implemented using the shadow file approach.
 * Upon top-level commit, modified pages are written to a shadow file,
 * referred to here as the "transaction file".
 * The mode of the shadow file is changed to indicate that it has been
 * committed.
 * NB:
 * A significant advantage of the shadow file approach is that disk writes
 * need not be atomic (assuming fsync() works).  The transaction file is
 * guaranteed to be complete when it is marked as committed and the recovery
 * procedure is idempotent (it can be reexecuted as many times as necessary
 * and a correct database will still result).
 *
 * The transaction file format is some initial number of directory pages, each
 * page TRANS_FILE_MIN_DIRSIZE bytes long, followed by the actual data.
 * Each directory page consists of pairs of integers.
 * The first six integers in the file are special:
 *     incarnation1, incarnation2,
 *     total # of pairs,
 *     transactionid1, transactionid2, transactionid3
 *
 * The first pair constitutes an incarnation number that matches the one in
 * the corresponding database's header.
 * The next integer is the total number of pairs of integers that follow.
 * The next three integers constitute the DbmTidName.
 * Of the remaining pairs, the first integer is the offset in the target dbm
 * file for the corresponding data and the second integer is the number of
 * bytes of data.  The data starts on the first page following the directory.
 *
 * Recovery occurs when the dbm file is first opened .  At this time, a check
 * is made to see if a transaction file already exists (its name is based on
 * that of the dbm file) and whether it has been committed.
 *
 * Consistency:
 * Traditional nested 2PL, as per J.E.B. Moss, is used.  Locks are released on
 * abort, whether the abort was caused by an error or a user request.
 *
 * Isolation:
 * Shadow pages are used to hide uncommitted updates from other transactions.
 *
 * Durability:
 * Top-level commits are written to a normal Unix file, except in the case
 * of volatile databases (which disappear when closed explicitly or
 * implicitly).
 *
 * Copyright 1992 Barry Brachman
 * See included file "copyright.h" for complete copyright information.
 */

/*
 * XXX
 * - Critical sections are not properly protected.
 * - Limited deadlock detection.
 * - A transaction over multiple dbm files has not be implemented.
 * - In the case of an indirect page, a lock may span a range of pages although
 *   it is explicit only on the first page.
 */

#include "local.h"

#include <sys/file.h>
#include <sys/stat.h>
#if _IBMR2 || _IBM
#define _BSD
#endif
#include <sys/dir.h>
#include <netdb.h>
#include <sys/param.h>
#include <fcntl.h>
#include <errno.h>

#ifdef mips
extern int errno;		/* MIPS bogosity. */
#endif

typedef struct {
	Tid			*tid;
	PhysPagenum pagenum;
} GrantTestArg;

#define RLOCK(X,A)	((A)->tid = TID(trans), (A)->pagenum = pagenum, \
					 ReadLock(X, grant_test, A))
#define WLOCK(X,A)	((A)->tid = TID(trans), (A)->pagenum = pagenum, \
					 WriteLock(X, grant_test, A))
#define UPLOCK(X,A)	((A)->tid = TID(trans), (A)->pagenum = pagenum, \
					 UpgradeLock(X, grant_test, A))

#ifdef THREADS
#define BEGIN_CRITICAL(L)	\
	if ((L < 0 && (L = NewSem(1)) < 0) \
		|| P(L) < 0) \
	return(DBM_TRANS_LOCK_REQUEST_FAILED)
#define END_CRITICAL(L)		V(L)
#else
#define BEGIN_CRITICAL(L)
#define END_CRITICAL(L)
#endif

#ifdef TRACING
#define TRACE(X)		fprintf X
#else
#define TRACE(X)
#endif

#define NOT_PRECOMMITTED		0
#define DISK_PRECOMMITTED		1
#define MEMORY_PRECOMMITTED		2

/* Transfile initial layout. */
#define TRANS_DIR_HEADER_LENGTH		6
#define TRANS_INCARNATION1(dir)		(dir[0])
#define TRANS_INCARNATION2(dir)		(dir[1])
#define TRANS_NPAIRS(dir)			(dir[2])
#define TRANS_TIDNAME1(dir)			(dir[3])
#define TRANS_TIDNAME2(dir)			(dir[4])
#define TRANS_TIDNAME3(dir)			(dir[5])
#define TRANS_PAIR_START(dir)		(dir + TRANS_DIR_HEADER_LENGTH)

extern long strtol PROTO((char *str, char **ptr, int base));
extern struct hostent *gethostbyname PROTO((char *name));

static DbmRc addActiveTrans PROTO((Tid *tid));
static DbmRc addLock PROTO((DbmTrans *trans, PhysPagenum pagenum,
							LockType mode, TransLock **tlp));
static DbmRc addPageLock PROTO((DbmTrans *trans, PhysPagenum pagenum,
						 DbfPageLock **plockp));
static DbmRc addTransLock PROTO((DbmTrans *trans, PhysPagenum pagenum,
						  DbfPageLock *plock, LockType mode, TransLock **tlp));
static DbmRc allocTid PROTO((Tid *parent, Tid **newtid));
static int cacheGet PROTO((Tid *tid, int flags, PageCache **cache_list));
static DbmRc checkWriteLock PROTO((DbmTrans *trans, PhysPagenum pagenum));
static int commit PROTO((Tid *tid, PageCache *cache_list));
static DbmRc commitLocks PROTO((Tid *tid));
static DbmRc commitTransFile PROTO((Tid *tid));
static DbmRc createTransFile PROTO((Tid *tid, int flags));
static int filesort PROTO((struct direct **pptr, struct direct **qptr));
static void freeRecovery PROTO((DbmRecovery *recovery_list));
static void freeTid PROTO((Tid *tid));
static void freeTransLock PROTO((TransLock *tl));
static DbmRc getLock PROTO((DbmTrans *trans, PhysPagenum pagenum, LockType mode,
					 TransLock **tlp));
static DbmRc getTransFiles PROTO((char *pathname, DbmRecovery **file_list));
static int grant_test PROTO((GrantTestArg *arg, LOCKMODE mode));
static DbmRc inheritLock PROTO((TransLock *ptl, TransLock *tl));
static int matcher PROTO((struct direct *d));
static DbmRc precommit PROTO((Tid *tid, int n, PageCache *cache_list));
static DbmRc releaseLock PROTO((TransLock *tl, DbmTerminateMode mode));
static DbmRc releaseLocks PROTO((Tid *tid, DbmTerminateMode mode));
static DbmRc reopenTransFile PROTO((Tid *tid, char *pathname));
static void removeActiveTrans PROTO((ActiveTrans *active));
static void removePageLock PROTO((DbmFile *dbf, DbfPageLock *pl));
static DbmRc removeTransFile PROTO((Tid *tid));
static void showDbfPageLock PROTO((DbfPageLock *li));

static ActiveTrans *active_transactions = NULL;
static int commit_semaph = -1;
static int excl_semaph = -1;

/*
 * Initiate a new transaction for a dbm.
 * If PARENT is NULL, this is a new top-level transaction.
 */
DbmRc
TransBegin PARAMS((parent, child),
Tid *parent			_AND_
Tid **child)
{
	Tid *tid;
	DbmRc rc;

#ifndef ALLOW_MULTIPLE_CHILDREN
	if (parent && parent->children)
		return(DBM_UNIMPLEMENTED);
#endif

	if (parent && parent->active->precommitted != NOT_PRECOMMITTED)
		return(DBM_TRANS_PRECOMMITTED);

	if ((rc = allocTid(parent, &tid)) != DBM_OK)
		return(rc);

	*child = tid;
	TRACE((stderr, "TransBegin\n"));
	return(DBM_OK);
}

/*
 * Read a page through the transaction mechanism.
 * Related transactions share a read lock, unrelated transactions must obtain
 * their own.
 */
DbmRc
TransRead PARAMS((trans, pagenum, nbytes, cache),
DbmTrans *trans				_AND_
PhysPagenum pagenum			_AND_
int nbytes					_AND_
PageCache **cache)
{
	DbmRc rc;
	Tid *tid;
	TransLock *tl;

	tid = TID(trans);
	if (!tid)
		return(DBM_UNIMPLEMENTED);
	if (tid->active->precommitted != NOT_PRECOMMITTED)
		return(DBM_TRANS_PRECOMMITTED);

	if (TID_DBF(tid) == NULL)
		TID_DBF(tid) = DBF(trans);
	else if (TID_DBF(tid) != DBF(trans))
		return(DBM_UNIMPLEMENTED);

	BEGIN_CRITICAL(excl_semaph);

	/* Obtain a read lock for the page. */
    if ((rc = getLock(trans, pagenum, TRANS_READLOCK, &tl)) != DBM_OK)
		goto done;

	/*
	 * Sanity check.  Note that a page may be used as part of an indirect
	 * entry, freed, and then reallocated for some other use all within the
	 * same transaction.  That's how a "deleted" page could be read.
	 */
	if (tl->cache && (tl->flags & CACHE_DELETED)) {
		fprintf(stderr, "TransRead: reading deleted page?\n");
		releaseLocks(tid, DBM_ABORT);
		rc = DBM_TRANS_LOCK_PROTOCOL_ERROR;
		goto done;
	}

	/*
	 * If the page was not obtained from a superior, get a cached copy or
	 * a copy from the dbm file.
	 */
	if (tl->cache == NULL || tl->cache->page == NULL) {
		/*
		 * Bug fix:
		 * The (indirect) page could have been deleted and is now being
		 * reassigned.
		 */
		if (tl->flags & CACHE_DELETED) {
			/* XXX Is it necessary to write lock PAGE_NEW cache entries? */
			rc = CacheRead(trans, pagenum, nbytes, &tl->cache);
			if (rc != DBM_OK)
				goto done;
			tl->flags &= (~CACHE_DELETED);
		}
		else {
			if (tl->lockinfo->cache) {
				tl->cache = tl->lockinfo->cache;
#ifdef NOTDEF
			fprintf(stderr, "TransRead1 %d: p=%d db=%s cache=%x\n", MyPid(),
				pagenum, tl->dbf->pathname, tl->lockinfo->cache);
#endif
			}
			else {
				/* XXX Is it necessary to write lock PAGE_NEW cache entries? */
				rc = CacheRead(trans, pagenum, nbytes, &tl->cache);
#ifdef NOTDEF
				fprintf(stderr, "TransRead2 %d: p=%d db=%s c=%x\n", MyPid(),
					pagenum, tl->dbf->pathname, tl->cache);
#endif
				if (rc != DBM_OK)
					goto done;
				tl->lockinfo->cache = tl->cache;
			}
			tl->flags |= CACHE_SHARED;
		}
		TRACE((stderr, "  TransRead:   %4d (base)\n", tl->lockinfo->pagenum));
	}
	else {
		tl->cache->flags &= (~PAGE_NEW);
		TRACE((stderr, "  TransRead:   %4d (shared)\n",
			   tl->lockinfo->pagenum));
	}

	*cache = tl->cache;
	rc = DBM_OK;

 done:
	END_CRITICAL(excl_semaph);

	return(rc);
}

/*
 * A page is about to be updated (modified).  Either a read lock must be
 * upgraded to a write lock or a write lock must be obtained.
 */
DbmRc
TransUpdate PARAMS((trans, cachep),
DbmTrans *trans			_AND_
PageCache **cachep)
{
	DbmFile *dbf;
	DbmRc rc;
	GrantTestArg grantarg;
	PageCache *cache;
	PhysPagenum pagenum;
	Tid *tid;
	TransLock *newtl, *tl;

	tid = TID(trans);
	if (!tid)
		return(DBM_UNIMPLEMENTED);
	if (tid->active->precommitted != NOT_PRECOMMITTED)
		return(DBM_TRANS_PRECOMMITTED);

	dbf = DBF(trans);
	if (TID_DBF(tid) == NULL)
		TID_DBF(tid) = dbf;
	else if (TID_DBF(tid) != dbf)
		return(DBM_UNIMPLEMENTED);

	BEGIN_CRITICAL(excl_semaph);

	cache = *cachep;
	pagenum = cache->pagenum;
    /* Search the transaction tree upwards for the page. */
    do {
        tl = tid->translock_hashtab[pagenum & TRANSLOCK_HASHTAB_MASK];
        while (tl && tl->lockinfo->pagenum != pagenum)
            tl = tl->next;
    } while (tl == NULL && (tid = tid->parent) != NULL);

	if (tl && tid == TID(trans)) {
		/* The transaction already holds or retains the lock. */
		if (tl->locktype == TRANS_WRITELOCK) {
			if (tl->flags & CACHE_DELETED) {
				/* The page was deleted and is being rewritten. */
				tl->flags &= (~CACHE_DELETED);
				tl->cache = cache;
			}
			TRACE((stderr, "  TransUpdate: %4d (nop)\n",
				   tl->lockinfo->pagenum));
			END_CRITICAL(excl_semaph);
			return(DBM_OK);
		}
		if (tl->locktype == TRANS_READLOCK) {
			if (UPLOCK(tl->lockinfo->lockid, &grantarg) < 0)
				return(DBM_TRANS_LOCK_REQUEST_FAILED);
			TRACE((stderr, "  TransUpdate: %4d (upgraded%s)\n",
				   tl->lockinfo->pagenum,
				   (tl->flags & CACHE_SHARED) ? ", dup" : ""));
		}
		else {
			/* Lock is retained. */
			if (WLOCK(tl->lockinfo->lockid, &grantarg) < 0)
				return(DBM_TRANS_LOCK_REQUEST_FAILED);
			TRACE((stderr, "  TransUpdate: %4d (retained%s)\n",
				   tl->lockinfo->pagenum,
				   (tl->flags & CACHE_SHARED) ? ", dup" : ""));
		}
		tl->locktype = TRANS_WRITELOCK;
		if (tl->flags & CACHE_SHARED) {
			if ((rc = CacheDup(cache, cachep)) != DBM_OK) {
				END_CRITICAL(excl_semaph);
				return(rc);
			}
			tl->cache = *cachep;
			tl->flags &= (~CACHE_SHARED);
		}

		END_CRITICAL(excl_semaph);
		return(DBM_OK);
	}

    if (tl) {
		/*
		 * A superior holds or retains the lock.
		 * If an inferior tries to get a write lock already held in any mode
		 * by a superior, a deadlock will result because the superior can't
		 * commit (and release the lock) but the inferior can't proceed.
		 * The child must be aborted.
		 */
		if (!TRANS_RETAINED_LOCK(tl->locktype)) {
			fprintf(stderr, "TransUpdate: deadlock detected!\n");
			if (tid->children) {
				fprintf(stderr, "TransUpdate: can't proceed!\n");
				END_CRITICAL(excl_semaph);
				return(DBM_UNIMPLEMENTED);
			}
			if ((rc = releaseLocks(tid, DBM_ABORT)) != DBM_OK) {
				END_CRITICAL(excl_semaph);
				return(rc);
			}
			freeTid(tid);
			END_CRITICAL(excl_semaph);
			return(DBM_TRANS_ABORTED);
		}

		/* A superior retains the lock. */
		if ((rc = addTransLock(trans, pagenum, tl->lockinfo,
							   TRANS_WRITELOCK, &newtl)) != DBM_OK) {
			END_CRITICAL(excl_semaph);
			return(rc);
		}
		if ((rc = CacheDup(cache, cachep)) != DBM_OK) {
			END_CRITICAL(excl_semaph);
			return(rc);
		}
		newtl->cache = *cachep;
		newtl->flags = CACHE_CLEAN;
		tl = newtl;

		TRACE((stderr, "  TransUpdate: %4d\n", tl->lockinfo->pagenum));
		END_CRITICAL(excl_semaph);
		return(DBM_OK);
	}

	/*
	 * When new pages are created (for example, in the case of indirect pages),
	 * no read lock exists so no upgrade is possible.
	 */
	if ((rc = addLock(trans, pagenum, TRANS_WRITELOCK, &tl)) != DBM_OK) {
		END_CRITICAL(excl_semaph);
		return(rc);
	}
	tl->cache = cache;

	TRACE((stderr, "  TransUpdate: %4d (new)\n", tl->lockinfo->pagenum));
	END_CRITICAL(excl_semaph);
	return(DBM_OK);
}

/*
 * Write a page through the transaction mechanism.
 */
DbmRc
TransWrite PARAMS((trans, cache),
DbmTrans *trans			_AND_
PageCache *cache)
{
	Tid *tid;
	DbmRc rc;

	tid = TID(trans);
	if (!tid)
		return(DBM_UNIMPLEMENTED);
	if (tid->active->precommitted != NOT_PRECOMMITTED)
		return(DBM_TRANS_PRECOMMITTED);

	if (TID_DBF(tid) == NULL)
		TID_DBF(tid) = DBF(trans);
	else if (TID_DBF(tid) != DBF(trans))
		return(DBM_UNIMPLEMENTED);

	if ((rc = checkWriteLock(trans, cache->pagenum)) != DBM_OK)
		return(rc);

	TRACE((stderr, "  TransWrite:  %4d\n", cache->pagenum));
	return(CacheWrite(trans, cache));
}

/*
 * Mark a page (or an implicit range of pages) as deleted and free the cached
 * page.
 */
DbmRc
TransDelete PARAMS((trans, pagenum),
DbmTrans *trans			_AND_
PhysPagenum pagenum)
{
	DbfPageLock *plock;
	DbmFile *dbf;
	DbmRc rc;
	Tid *tid;
	TransLock *tl;

	tid = TID(trans);
	if (tid == NULL)
		return(DBM_UNIMPLEMENTED);
	if (tid->active->precommitted != NOT_PRECOMMITTED)
		return(DBM_TRANS_PRECOMMITTED);

	dbf = DBF(trans);
	if (TID_DBF(tid) == NULL)
		TID_DBF(tid) = dbf;
	else if (TID_DBF(tid) != dbf)
		return(DBM_UNIMPLEMENTED);

	tl = tid->translock_hashtab[pagenum & TRANSLOCK_HASHTAB_MASK];
	while (tl && tl->lockinfo->pagenum != pagenum)
		tl = tl->next;

	/*
	 * A few words about the following weirdness.  In the case of indirect
	 * pages, it's possible to ask for them to be deleted although they
	 * haven't been read (i.e., tl == NULL) by this transaction.  In this case,
	 * the dbm should be volatile and a bogus TransLock will be created to
	 * record the deletion.
	 */
	if (tl) {
		TRACE((stderr, "  TransDelete: %4d\n", tl->lockinfo->pagenum));
		CacheFree(tl->cache);
		tl->cache = NULL;
		tl->flags |= CACHE_DELETED;
		return(DBM_OK);
	}
	else if (!dbf->persistent) {
		/* Bug Fix:
		 * If this is a persistent dbm then just marking the pages as free in
		 * DeleteIndirect() is sufficient.  If the dbm is volatile, however,
		 * the base copy must also be freed.
		 * The trick is to write lock the page(s) here and mark them as
		 * deleted.  At top-level commit time they can be discarded.
		 */
		BEGIN_CRITICAL(excl_semaph);

		plock = dbf->lock_hashtab[pagenum & LOCK_HASHTAB_MASK];
		while (plock && plock->pagenum != pagenum)
			plock = plock->next;
		if (plock == NULL)
			return(DBM_PAGE_NOT_FOUND);
		rc = addTransLock(trans, pagenum, plock, TRANS_WRITELOCK, &tl);
		if (rc == DBM_OK)
			tl->flags = CACHE_DELETED;

		END_CRITICAL(excl_semaph);
		return(rc);
	}

	return(DBM_PAGE_NOT_FOUND);
}

/*
 * Cancel the transaction, free resources.
 *
 * XXX should abort all children too.
 */
DbmRc
TransAbort PARAMS((tid),
Tid *tid)
{
	DbmFile *dbf;
	DbmRc rc;

	if (tid == NULL)
		return(DBM_OK);

	if (tid->children) {
		/* XXX All children must eventually be aborted. */
		return(DBM_UNIMPLEMENTED);
	}

	BEGIN_CRITICAL(excl_semaph);

	rc = DBM_OK;
	dbf = TID_DBF(tid);

	if (dbf) {
		rc = DBM_OK;
		if (tid->active->precommitted != NOT_PRECOMMITTED)
			rc = removeTransFile(tid);

		if (rc != DBM_OK || (rc = releaseLocks(tid, DBM_ABORT)) != DBM_OK)
			goto done_abort;
	}

	freeTid(tid);

 done_abort:
	TRACE((stderr, "TransAbort\n"));
	END_CRITICAL(excl_semaph);

	return(rc);
}

DbmRc
TransTerminate PARAMS((dbm, mode),
Dbm *dbm					_AND_
DbmTerminateMode mode)
{
	ActiveTrans *active;
	DbmRc rc;

	for (active = active_transactions; active; active = active->next) {
		if (active->dbm_id == dbm->id) {
			if (mode == DBM_ABORT)
				rc = TransAbort(active->rootTid);
			else if (mode == DBM_COMMIT)
				rc = TransCommit(active->rootTid);
			else
				rc = DBM_INVALID_ARGUMENT;
			if (rc != DBM_OK)
				return(rc);
		}
	}
	TRACE((stderr, "TransTerminate\n"));
	return(DBM_OK);
}

/*
 * Free any top-level cached pages that are not in use.
 * This is called by DbmClose() when aborting and closing a dbm.
 */
DbmRc
TransUncache PARAMS((dbf),
DbmFile *dbf)
{
	int i;
	DbfPageLock *next, *pl;

	BEGIN_CRITICAL(excl_semaph);

	for (i = 0; i < LOCK_HASHTAB_SIZE; i++) {
		pl = dbf->lock_hashtab[i];
		while (pl) {
			next = pl->next;
			if (pl->nheld + pl->read_retained + pl->write_retained == 0)
				removePageLock(dbf, pl);
			pl = next;
		}
	}

	TRACE((stderr, "  TransUncache\n"));
	END_CRITICAL(excl_semaph);
	return(DBM_OK);
}

/*
 * Prepare for a top-level atomic commit.
 * This might be used by a two-phase commit protocol, for example.
 * The transaction is marked so that no operations except abort or commit
 * are possible.
 * If volatile, there's not much to do.
 * Otherwise, the transaction file must be created but not marked as committed.
 * All locks must be maintained.
 */
DbmRc
TransPrecommit PARAMS((tid, tidname),
Tid *tid		_AND_
TidName *tidname)
{
	int n;
	DbmFile *dbf;
	DbmRc rc;
	PageCache *cache_list;

 	if (tid == NULL)
		return(DBM_UNIMPLEMENTED);
	if (tid->active->precommitted != NOT_PRECOMMITTED)
		return(DBM_TRANS_PRECOMMITTED);		/* DBM_OK? */
	if (tid->parent)
		return(DBM_TRANS_NOT_TOPLEVEL);

	if (tid->children) {
		/*
		 * XXX There are uncommitted children.  We should block until they
		 * are resolved.
		 */
		return(DBM_TRANS_IN_PROGRESS);
	}

	rc = DBM_OK;

	dbf = TID_DBF(tid);
	if (dbf == NULL)
		goto done_precommit;

	if (!dbf->persistent) {
		tid->active->precommitted = MEMORY_PRECOMMITTED;
		goto done_precommit;
	}

	/* XXX Should allow simultaneous commits. */
	BEGIN_CRITICAL(excl_semaph);

	/* If there are no dirty pages, then not much to do. */
	if ((n = cacheGet(tid, PAGE_DIRTY, &cache_list)) > 0)
		rc = precommit(tid, n, cache_list);

	END_CRITICAL(excl_semaph);

 done_precommit:

	if (rc == DBM_OK) {
		if (tidname) {
			tidname->hostid = tid->active->tidname.hostid;
			tidname->start_time = tid->active->tidname.start_time;
			tidname->count = tid->active->tidname.count;
		}
		TRACE((stderr, "TransPrecommit\n"));
	}
	return(rc);
}

/*
 * Create the transaction file and begin committing it.
 * Free resources.
 * XXX Check if file truncation possible.
 */
DbmRc
TransCommit PARAMS((tid),
Tid *tid)
{
	int n;
	DbmFile *dbf;
	DbmRc rc;
	PageCache *cache_list;

 	if (tid == NULL) {
		TRACE((stderr, "TransCommit\n"));
		return(DBM_OK);
	}

	/* XXX If a superior has aborted, we must abort here. */

	if (tid->children) {
		/*
		 * XXX There are uncommitted children.  We should block until they
		 * are resolved.
		 */
		return(DBM_TRANS_IN_PROGRESS);
	}

	dbf = TID_DBF(tid);
	if (dbf == NULL) {
		freeTid(tid);
		TRACE((stderr, "TransCommit\n"));
		return(DBM_OK);
	}

	/*
	 * If this is not a top-level commit, just shuffle locks.
	 */
	if (tid->parent) {
		commitLocks(tid);
		freeTid(tid);
		TRACE((stderr, "TransCommit\n"));
		return(DBM_OK);
	}

	/* Top-level commit */

	/* XXX Must be atomic. */
	if (!dbf->persistent) {
		releaseLocks(tid, DBM_COMMIT);
		freeTid(tid);
		TRACE((stderr, "TransCommit\n"));
		return(DBM_OK);
	}

	/* XXX Should allow simultaneous commits. */
	/* BEGIN_CRITICAL(commit_semaph); */
	BEGIN_CRITICAL(excl_semaph);

	if (tid->active->precommitted == NOT_PRECOMMITTED) {
		/* If there are no dirty pages, then just clean up. */
		if ((n = cacheGet(tid, PAGE_DIRTY, &cache_list)) == 0) {
			releaseLocks(tid, DBM_COMMIT);
			freeTid(tid);
			/* END_CRITICAL(commit_semaph); */
			END_CRITICAL(excl_semaph);
			TRACE((stderr, "TransCommit\n"));
			return(DBM_OK);
		}

		if ((rc = precommit(tid, n, cache_list)) != DBM_OK)
			goto done_commit;
	}

	/* Commit! */
	if (fchmod(TID_FD(tid), TRANS_FILE_COMMIT_MODE) < 0)
		goto done_commit;
	if (fsync(TID_FD(tid)) < 0)
		goto done_commit;

	rc = commitTransFile(tid);

 done_commit:

	END_CRITICAL(excl_semaph);
	/* END_CRITICAL(commit_semaph); */

	releaseLocks(tid, rc == DBM_OK ? DBM_COMMIT : DBM_ABORT);   /* XXX */

	freeTid(tid);
	if (rc == DBM_OK)
		TRACE((stderr, "TransCommit\n"));
	return(rc);
}

/*
 * This is a top-level precommit on a persistent dbm.
 * Build the transaction file but don't commit it.
 * Info about the transaction file will be maintained in the ActiveTrans.
 * Assumes all arguments are valid.
 */
static DbmRc
precommit PARAMS((tid, n, cache_list),
Tid *tid					_AND_
int n						_AND_
PageCache *cache_list)
{
	int fd, i, need;
	int *dirp, *transdir;
	off_t offset;
	DbmFile *dbf;
	DbmRc rc;
	PageCache *cache;

	transdir = NULL;
	if ((rc = createTransFile(tid, O_RDWR | O_CREAT | O_EXCL)) != DBM_OK)
		goto done_precommit;

	/*
	 * Now go through the list of dirty pages for this transaction,
	 * write them to the transaction file.
	 */
	need = (TRANS_DIR_HEADER_LENGTH + (n * 2)) * sizeof(int);
	need = ((need + TRANS_FILE_MIN_DIRSIZE - 1)
			/ TRANS_FILE_MIN_DIRSIZE) * TRANS_FILE_MIN_DIRSIZE;
	if ((dirp = transdir = (int *) NAMEDMALLOC("Transdir", need)) == NULL) {
		rc = DBM_MALLOC_FAILED;
		goto done_precommit;
	}

	dbf = TID_DBF(tid);
	fd = TID_FD(tid);

	TRANS_INCARNATION1(dirp) = dbf->incarnation_sec;
	TRANS_INCARNATION2(dirp) = dbf->incarnation_usec;
	TRANS_NPAIRS(dirp) = n;
	TRANS_TIDNAME1(dirp) = tid->active->tidname.hostid;
	TRANS_TIDNAME2(dirp) = tid->active->tidname.start_time;
	TRANS_TIDNAME3(dirp) = tid->active->tidname.count;
	dirp = TRANS_PAIR_START(dirp);

	rc = DBM_TRANS_IO_ERROR;

	offset = (off_t) need;
	cache = cache_list;
	for (i = 0; i < n; i++) {
		*dirp++ = cache->pagenum * dbf->allocsize;
		*dirp++ = cache->nbytes;
		if (SeekNWrite(fd, (char *) cache->page, cache->nbytes, offset,
					   SEEK_SET)
			!= cache->nbytes)
			goto done_precommit;
		offset += cache->nbytes;
		cache = cache->next;
	}

	if (SeekNWrite(fd, (char *) transdir, need, (off_t) 0, SEEK_SET) != need)
		goto done_precommit;

	/*
	 * First make sure the data is secured to disk, then make sure it
	 * is marked committed before starting to overlay.
	 */
	if (fsync(fd) < 0)
		goto done_precommit;

	if (fchmod(fd, TRANS_FILE_PRECOMMIT_MODE) < 0)
		goto done_precommit;
	if (fsync(fd) < 0)
		goto done_precommit;

	rc = DBM_OK;
	tid->active->precommitted = DISK_PRECOMMITTED;

 done_precommit:
	if (transdir)
		FREE(transdir);
	return(rc);
}

/*
 * Create a list of cache entries for TID that satisfies the given FLAGS
 * (ignore condition if FLAGS is zero).
 * Return the length of the list.
 */
static int
cacheGet PARAMS((tid, flags, cache_list),
Tid *tid					_AND_
int flags					_AND_
PageCache **cache_list)
{
	int i, n;
	PageCache *cache;
	TransLock *tl;

	*cache_list = cache = NULL;
	n = 0;
	for (i = 0; i < TRANSLOCK_HASHTAB_SIZE; i++) {
		tl = tid->translock_hashtab[i];
		while (tl) {
			if (!flags || (tl->cache && (tl->cache->flags & flags))) {
				if (cache)
					cache = cache->next = tl->cache;
				else
					*cache_list = cache = tl->cache;
				if (cache->page == NULL)
					fprintf(stderr, "cacheGet: page %d is empty!\n",
							cache->pagenum);
				n++;
			}
			tl = tl->next;
		}
	}
	if (cache)
		cache->next = NULL;
	return(n);
}

/*
 * A user is trying to recover a precommitted transaction.
 * Return a new transaction identifier.
 * Update the list by removing the just-processed item.
 */
DbmRc
TransUserRecover PARAMS((dbf, recoveryp, tidp),
DbmFile *dbf			_AND_
DbmRecovery **recoveryp	_AND_
Tid **tidp)
{
	char *pathname;
	DbmRc rc;
	DbmRecovery *recovery;
	Tid *tid;

	if (recoveryp == NULL || (recovery = *recoveryp) == NULL)
		return(DBM_OK);

	if ((rc = allocTid(NULL, tidp)) != DBM_OK)
		return(rc);
	tid = *tidp;
	TID_DBF(tid) = dbf;

	/* XXX Overwrite bogus new tidname info. */
	tid->active->tidname.hostid = recovery->tidname.hostid;
	tid->active->tidname.start_time = recovery->tidname.start_time;
	tid->active->tidname.count = recovery->tidname.count;

	/* XXX Should mark dbf as in recovery phase to disallow normal operation */

	if ((rc = reopenTransFile(tid, recovery->pathname)) != DBM_OK) {
		freeTid(tid);
		*tidp = NULL;
		/* XXX user is now stuck since other recovery names not accessible */
		return(rc);
	}

	*recoveryp = recovery->next;
	FREE(recovery->pathname);
	FREE(recovery);

	tid->active->precommitted = DISK_PRECOMMITTED;
	return(DBM_OK);
}

/*
 * This is called when a dbm is opened for the first time.
 * It searches for any transactions that are around if the
 * dbm crashed the last time it was used.
 *
 * Check for transaction files:
 * 1. Commit each one marked as committed
 * 2. If requested, return a list of those marked precommitted
 * 3. Unlink all others
 */
DbmRc
TransRecover PARAMS((dbm, rl),
Dbm *dbm			_AND_
DbmRecovery **rl)
{
	int fd, len, n;
	struct stat statbuf;
	DbmFile *dbf;
	DbmRc rc;
	DbmRecovery *keep, *keep_list, *next, *recovery, *recovery_list;
	static int recover_semaph = -1;

	BEGIN_CRITICAL(recover_semaph);

	dbf = dbm->dbmfile;

	rc = getTransFiles(dbf->pathname, &recovery_list);

	keep_list = keep = NULL;
	recovery = recovery_list;
	while (recovery) {
		if ((fd = open(recovery->pathname, O_RDONLY, 0)) < 0) {
			rc = DBM_RECOVER_FAILED;
			goto done;
		}

		if (fstat(fd, &statbuf) < 0
			|| ((statbuf.st_mode & S_IFMT) != S_IFREG)) {
			freeRecovery(recovery_list);
			rc = DBM_RECOVER_FAILED;
			goto done;
		}

		if (rl && (statbuf.st_mode & 0777) == TRANS_FILE_PRECOMMIT_MODE) {
			int *dirp;
			char buf[TRANS_DIR_HEADER_LENGTH * sizeof(int)];

			len = TRANS_DIR_HEADER_LENGTH * sizeof(int);
			if ((n = read(fd, buf, len)) != len) {
				rc = DBM_RECOVER_FAILED;
				goto done;
			}

			if (keep) {
				keep->next = recovery;
				keep = recovery;
			}
			else
				keep_list = keep = recovery;
			/* Update pointer first! */
			recovery = recovery->next;
			dirp = (int *) buf;
			keep->tidname.hostid = TRANS_TIDNAME1(dirp);
			keep->tidname.start_time = TRANS_TIDNAME2(dirp);
			keep->tidname.count = TRANS_TIDNAME3(dirp);
			keep->next = NULL;
			close(fd);
			continue;
		}

		else if ((statbuf.st_mode & 0777) == TRANS_FILE_COMMIT_MODE) {
			Tid *tid;

			if ((rc = allocTid(NULL, &tid)) != DBM_OK)
				goto done;
			TID_DBF(tid) = dbf;

			if ((rc = reopenTransFile(tid, recovery->pathname)) != DBM_OK) {
				freeTid(tid);
				goto done;
			}

			if ((rc = commitTransFile(tid)) != DBM_OK) {
				freeRecovery(recovery_list);
				freeTid(tid);
				goto done;
			}
			freeTid(tid);
		}
		else
			/*
			 * If the caller isn't interested, precommitted transactions are
			 * lost.
			 */
			unlink(recovery->pathname);

		next = recovery->next;
		FREE(recovery);
		recovery = next;
		close(fd);
	}

	fd = -1;
	if (rl)
		*rl = keep_list;

	rc = DBM_OK;

 done:
	END_CRITICAL(recover_semaph);

	if (fd >= 0)
		close(fd);

	return(rc);
}

static void
freeRecovery PARAMS((recovery_list),
DbmRecovery *recovery_list)
{
	DbmRecovery *next;

	while (recovery_list) {
		next = recovery_list->next;
		FREE(recovery_list->pathname);
		FREE(recovery_list);
		recovery_list = next;
	}
}

/*
 * Commit a transaction file as part of recovery.
 */
static DbmRc
commitTransFile PARAMS((tid),
Tid *tid)
{
	int bufsize, *dirp, fd, i, n, need, *transdir;
	DbmFile *dbf;
	char *p, *pathname, *transbuf;
	DbmRc rc;

	/*
	 * Read in the header.
	 * This may be done in two steps if the header is large.
	 */
	dbf = TID_DBF(tid);
	fd = TID_FD(tid);
	if ((dirp = transdir =
		 (int *) NAMEDMALLOC("Transdir", TRANS_FILE_MIN_DIRSIZE)) == NULL) {
		rc = DBM_MALLOC_FAILED;
		goto cleanup;
	}
	rc = DBM_RECOVER_IO_ERROR;

	if (SeekNRead(fd, (char *) dirp, TRANS_FILE_MIN_DIRSIZE, (off_t) 0,
		SEEK_SET) != TRANS_FILE_MIN_DIRSIZE)
		goto cleanup;

	/* See precommit(). */
	if (TRANS_INCARNATION1(dirp) != dbf->incarnation_sec
		|| TRANS_INCARNATION2(dirp) != dbf->incarnation_usec) {
		rc = DBM_BAD_RECOVER_FILE;
		goto cleanup;
	}

	n = TRANS_NPAIRS(dirp);
	need = (TRANS_DIR_HEADER_LENGTH + (n * 2)) * sizeof(int);
	need = ((need + TRANS_FILE_MIN_DIRSIZE - 1)
			/ TRANS_FILE_MIN_DIRSIZE) * TRANS_FILE_MIN_DIRSIZE;
	if (need > TRANS_FILE_MIN_DIRSIZE) {
		if ((transdir = (int *) REALLOC(dirp, need)) == NULL) {
			/* dirp still points to valid memory. */
			FREE(dirp);
			rc = DBM_MALLOC_FAILED;
			goto cleanup;
		}
		dirp = transdir;
		if (READ(fd, (char *) (dirp + TRANS_FILE_MIN_DIRSIZE),
				 need - TRANS_FILE_MIN_DIRSIZE)
			!= (need - TRANS_FILE_MIN_DIRSIZE))
			goto cleanup;
	}

	/*
	 * Now read from the transaction file and write to the dbm file.
	 */
	dirp = TRANS_PAIR_START(dirp);
	transbuf = NULL;
	bufsize = 0;
	for (i = 0; i < n; i++) {
		if (bufsize < dirp[1]) {
			bufsize = dirp[1];
			if (transbuf) {
				if ((p = (char *) REALLOC(transbuf, bufsize)) == NULL) {
					FREE(transbuf);
					rc = DBM_MALLOC_FAILED;
					goto cleanup;
				}
			}
			else {
				if ((p = (char *) NAMEDMALLOC("Transbuf", bufsize)) == NULL) {
					rc = DBM_MALLOC_FAILED;
					goto cleanup;
				}
			}
			transbuf = p;
		}
		if (READ(fd, transbuf, dirp[1]) != dirp[1])
			goto cleanup;
		if (SeekNWrite(dbf->fd, (char *) transbuf, dirp[1], (off_t) dirp[0],
					   SEEK_SET) != dirp[1]) {
			if (errno == ENOSPC)
				rc = DBM_RECOVER_NO_DISK_SPACE;
			goto cleanup;
		}
		dirp += 2;
	}

	if (fsync(dbf->fd) == 0)
		rc = DBM_OK;

 cleanup:
	if (rc == DBM_OK)
		rc = removeTransFile(tid);
	if (transdir)
		FREE(transdir);
	if (transbuf)
		FREE(transbuf);

	return(rc);
}

/*
 * Execute a commit, writing pages on top of the dbm file.
 * Makes sure the data has been written to disk before returning.
 *
 * Note that other I/O operations may be concurrent.
 */
static int
commit PARAMS((tid, cache_list),
Tid *tid					_AND_
PageCache *cache_list)
{
	off_t offset;
	DbmFile *dbf;
	PageCache *cache;

	dbf = TID_DBF(tid);

	for (cache = cache_list; cache != NULL; cache = cache->next) {
		offset = (off_t) (cache->pagenum * dbf->allocsize);
		if (SeekNWrite(dbf->fd, cache->page, cache->nbytes, offset, SEEK_SET)
			!= cache->nbytes)
			return(-1);
	}

	return(fsync(dbf->fd));
}

/*
 * Get a lock of the given MODE for the given PAGENUM.
 *
 * A lock may have to be created and then obtained.
 * Once obtained, lock status information must be associated with both
 * the transaction (for efficient lock handling at commit/abort time) and the
 * dbm (to keep track of which pages are locked).
 */
static DbmRc
getLock PARAMS((trans, pagenum, mode, tlp),
DbmTrans *trans				_AND_
PhysPagenum pagenum			_AND_
LockType mode				_AND_
TransLock **tlp)
{
	int st;
	DbfPageLock *pl;
	DbmRc rc;
	GrantTestArg grantarg;
	Tid *tid;
	TransLock *tl;

	tid = TID(trans);

	/* Search the transaction tree upwards for the page. */
	do {
		tl = tid->translock_hashtab[pagenum & TRANSLOCK_HASHTAB_MASK];
		while (tl && tl->lockinfo->pagenum != pagenum)
			tl = tl->next;
	} while (tl == NULL && (tid = tid->parent) != NULL);

	if (tl && tid == TID(trans)) {
		/*
		 * It's already held or retained by this transaction.
		 */
		if (tl->locktype == TRANS_READLOCK && mode == TRANS_WRITELOCK) {
			fprintf(stderr, "getLock: TransUpdate required!\n");
			return(DBM_TRANS_LOCK_PROTOCOL_ERROR);
		}
		/*
		 * XXX These locking rules allow a transaction to acquire a lock that
		 * it retains.  See grant_test().
		 */
		if (TRANS_RETAINED_LOCK(tl->locktype)) {
			pl = tl->lockinfo;
			st = (mode == TRANS_READLOCK
				  ? RLOCK(pl->lockid, &grantarg)
				  : WLOCK(pl->lockid, &grantarg));
			if (st < 0)
				return(DBM_TRANS_LOCK_REQUEST_FAILED);
			if (tl->locktype == TRANS_READLOCK_RETAINED)
				pl->read_retained--;
			else
				pl->write_retained--;
			tl->locktype = mode;
			pl->nheld++;
		}

		*tlp = tl;
		return(DBM_OK);
	}

	if (tl) {
		/* The lock was found in a superior transaction. */
		pl = tl->lockinfo;
		if ((rc = addTransLock(trans, pagenum, pl, mode, tlp)) != DBM_OK)
			return(rc);
		if (mode == TRANS_READLOCK) {
			(*tlp)->cache = tl->cache;
			(*tlp)->flags |= CACHE_SHARED;
		}
		return(DBM_OK);
	}

	return(addLock(trans, pagenum, mode, tlp));
}

/*
 * A subtransaction is committing, so merge all of its locks with its parent's.
 * This may involve merging duplicate locks into the higher of the two types.
 */
static DbmRc
commitLocks PARAMS((tid),
Tid *tid)
{
	int i;
	DbfPageLock *plock;
	DbmRc rc;
	PhysPagenum pagenum;
	Tid *parent;
	TransLock *next, *ptl, **ptlp, *tl, **tlp;

	if (tid->children) {
		fprintf(stderr, "commitLocks: transaction has child!\n");
		return(DBM_TRANS_LOCK_PROTOCOL_ERROR);
	}

	BEGIN_CRITICAL(excl_semaph);

	parent = tid->parent;
	tlp = tid->translock_hashtab;
	for (i = 0; i < TRANSLOCK_HASHTAB_SIZE; i++) {
		if ((tl = *tlp) == NULL) {
			tlp++;
			continue;
		}
		pagenum = tl->lockinfo->pagenum;
		ptlp = &parent->translock_hashtab[pagenum & TRANSLOCK_HASHTAB_MASK];

		/* For each page in the bucket, search the parent for the page. */
		while (tl) {
			plock = tl->lockinfo;
			pagenum = plock->pagenum;
			ptl = *ptlp;
			while (ptl && ptl->lockinfo->pagenum != pagenum)
				ptl = ptl->next;
			next = tl->next;
			if (ptl) {
				/*
				 * Parent has associated state. If the child's data is
				 * different, the newer version must replace the older.
				 */
				if (ptl->cache != tl->cache)
					CacheFree(ptl->cache);
				ptl->flags = tl->flags;
				ptl->cache = tl->cache;

				if ((rc = inheritLock(ptl, tl)) != DBM_OK) {
					END_CRITICAL(excl_semaph);
					return(rc);
				}
				freeTransLock(tl);
			}
			else {
				/* Parent must inherit the state. */
				tl->next = *ptlp;
				*ptlp = tl;
				if (tl->locktype == TRANS_READLOCK) {
					tl->locktype = TRANS_READLOCK_RETAINED;
					plock->read_retained++;
					plock->nheld--;
					Unlock(plock->lockid);
				}
				else if (tl->locktype == TRANS_WRITELOCK) {
					tl->locktype = TRANS_WRITELOCK_RETAINED;
					plock->write_retained++;
					plock->nheld--;
					Unlock(plock->lockid);
				}
			}
			tl = next;
		}
		*tlp++ = NULL;
	}

	END_CRITICAL(excl_semaph);
	return(DBM_OK);
}

static DbmRc
inheritLock PARAMS((ptl, tl),
TransLock *ptl			_AND_
TransLock *tl)
{
	DbfPageLock *plock;

	plock = ptl->lockinfo;

	switch (tl->locktype) {
	case TRANS_READLOCK:
		if (ptl->locktype == TRANS_WRITELOCK) {
			fprintf(stderr, "inheritLock: impossible readlock state.\n");
			return(DBM_TRANS_LOCK_PROTOCOL_ERROR);
		}
		plock->nheld--;
		/* Parent's locktype doesn't change. */
		Unlock(plock->lockid);
		break;

	case TRANS_READLOCK_RETAINED:
		plock->read_retained--;
		/* Parent's locktype doesn't change. */
		break;

	case TRANS_WRITELOCK:
		if (!TRANS_RETAINED_LOCK(ptl->locktype)) {
			fprintf(stderr, "inheritLock: impossible writelock state.\n");
			return(DBM_TRANS_LOCK_PROTOCOL_ERROR);
		}
		if (ptl->locktype == TRANS_READLOCK_RETAINED) {
			plock->read_retained--;
			plock->write_retained++;
		}
		plock->nheld--;
		ptl->locktype = TRANS_WRITELOCK_RETAINED;
		Unlock(plock->lockid);
		break;

	case TRANS_WRITELOCK_RETAINED:
		if (!TRANS_RETAINED_LOCK(ptl->locktype)) {
			fprintf(stderr, "inheritLock: impossible write_retained state.\n");
			return(DBM_TRANS_LOCK_PROTOCOL_ERROR);
		}
		if (ptl->locktype == TRANS_READLOCK_RETAINED)
			plock->read_retained--;
		else
			plock->write_retained--;
		ptl->locktype = TRANS_WRITELOCK_RETAINED;
		break;

	default:
		fprintf(stderr, "inheritLock: invalid locktype?\n");
		return(DBM_TRANS_LOCK_PROTOCOL_ERROR);
		/*NOTREACHED*/
	}
	return(DBM_OK);
}


/*
 * Create a per-dbm lock record for PAGENUM (may reuse a cached record) and
 * make a per-transaction lock record.
 */
static DbmRc
addLock PARAMS((trans, pagenum, mode, tlp),
DbmTrans *trans				_AND_
PhysPagenum pagenum			_AND_
LockType mode				_AND_
TransLock **tlp)
{
	int did_create;
	DbfPageLock *plock;
	DbmFile *dbf;
	DbmRc rc;

	dbf = DBF(trans);

	/* BEGIN CRITICAL SECTION */

	plock = dbf->lock_hashtab[pagenum & LOCK_HASHTAB_MASK];
	while (plock && plock->pagenum != pagenum)
		plock = plock->next;

	did_create = 0;
	if (plock == NULL) {
		/* Create a new lock. */
		if ((rc = addPageLock(trans, pagenum, &plock)) != DBM_OK)
			return(rc);
		did_create = 1;
	}

	if ((rc = addTransLock(trans, pagenum, plock, mode, tlp)) != DBM_OK) {
		if (did_create)
			removePageLock(dbf, plock);
		return(rc);
	}

	return(DBM_OK);
}

/*
 * Deallocate locks belonging to a transaction, either because of a top-level
 * commit or an abort at any level.
 */
static DbmRc
releaseLocks PARAMS((tid, mode),
Tid *tid					_AND_
DbmTerminateMode mode)
{
	int i;
	DbmRc rc;
	TransLock *next, *tl;

	if (tid->children) {
		fprintf(stderr, "releaseLocks: transaction has child!\n");
		return(DBM_TRANS_LOCK_PROTOCOL_ERROR);
	}

	/* BEGIN_CRITICAL(excl_semaph); */

	/*
	 * For each lock held but no longer referenced, release it.
	 */
	for (i = 0; i < TRANSLOCK_HASHTAB_SIZE; i++) {
		tl = tid->translock_hashtab[i];
		while (tl) {
			next = tl->next;
			if ((rc = releaseLock(tl, mode)) != DBM_OK) {
				/* END_CRITICAL(excl_semaph); */
				return(rc);
			}
			tl = next;
		}
		tid->translock_hashtab[i] = NULL;
	}

	/* END_CRITICAL(excl_semaph); */
	return(DBM_OK);
}

/*
 * Free a transaction's lock, cleaning up the global lock on the page if this
 * is the last reference.
 * This is due to a top-level commit or an abort at any level.
 * XXX Must only be called from releaseLocks() since lists aren't adjusted.
 */
static DbmRc
releaseLock PARAMS((tl, mode),
TransLock *tl				_AND_
DbmTerminateMode mode)
{
	int do_freelock, do_unlock, lockid;
	DbfPageLock *pl;

	pl = tl->lockinfo;

	if (mode == DBM_ABORT) {
		if (tl->cache) {
			if (tl->cache->flags & PAGE_DIRTY) {
				/* If the page is shared at the top, zap the reference. */
				if (pl->cache == tl->cache)
					pl->cache = NULL;
				CacheFree(tl->cache);
				tl->cache = NULL;
			}
		}
	}
	else {
		if (tl->cache)
			tl->cache->flags = PAGE_CLEAN;
		/*
		 * Note that the following handles the special case where the page
		 * has been deleted since tl->cache will be NULL but not pl->cache.
		 */
		if (pl->cache != tl->cache) {
			if (pl->cache) {
				CacheFree(pl->cache);
			}
			pl->cache = tl->cache;
		}
	}

	do_unlock = 0;
	switch (tl->locktype) {
	case TRANS_READLOCK:
	case TRANS_WRITELOCK:
		pl->nheld--;
		do_unlock = 1;
		break;

	case TRANS_READLOCK_RETAINED:
		pl->read_retained--;
		break;

	case TRANS_WRITELOCK_RETAINED:
		pl->write_retained--;
		break;

	default:
		fprintf(stderr, "releaseLock: invalid locktype!\n");
		return(DBM_TRANS_LOCK_PROTOCOL_ERROR);
		/*NOTREACHED*/
	}

	lockid = pl->lockid;
	do_freelock = 0;
	if (pl->nheld + pl->read_retained + pl->write_retained == 0) {
		/* The page is no longer in use. */
		if (tl->dbf->persistent) {
			/* XXX Some caching could be done here when not aborting. */
			removePageLock(tl->dbf, pl);
		}
		else if ((mode == DBM_COMMIT && (tl->flags & CACHE_DELETED)) ||
				 (mode == DBM_ABORT
				  && pl->cache && (pl->cache->flags & PAGE_DIRTY))) {
			removePageLock(tl->dbf, pl);
		}
		else
			pl->lockid = -1;	/* Note that it's invalid. */
		do_freelock = 1;
	}

	freeTransLock(tl);
	if (do_unlock)
		Unlock(lockid);
	if (do_freelock)
		FreeLock(lockid);

	return(DBM_OK);
}

/*
 * Sanity check.  The transaction must have gotten a write lock by now.
 */
static DbmRc
checkWriteLock PARAMS((trans, pagenum),
DbmTrans *trans				_AND_
PhysPagenum pagenum)
{
	Tid *tid;
	TransLock *tl;

	tid = TID(trans);
    /* Search the transaction tree upwards for the page. */
    do {
        tl = tid->translock_hashtab[pagenum & TRANSLOCK_HASHTAB_MASK];
        while (tl && tl->lockinfo->pagenum != pagenum)
            tl = tl->next;
    } while (tl == NULL && (tid = tid->parent) != NULL);

	if (tid == TID(trans) && tl->locktype == TRANS_WRITELOCK) {
		/* Already locked by this transaction. */
		return(DBM_OK);
	}

	fprintf(stderr, "checkWriteLock: no write lock!\n");
	return(DBM_TRANS_LOCK_PROTOCOL_ERROR);
}

static DbmRc
addTransLock PARAMS((trans, pagenum, pl, mode, tlp),
DbmTrans *trans				_AND_
PhysPagenum pagenum			_AND_
DbfPageLock *pl				_AND_
LockType mode				_AND_
TransLock **tlp)
{
	int st;
	GrantTestArg grantarg;
	TransLock **head, *tl;

    if ((tl = (TransLock *) NAMEDMALLOC("TransLock", sizeof(TransLock)))
		== NULL)
		return(DBM_MALLOC_FAILED);

    tl->dbf = DBF(trans);
    tl->lockinfo = pl;
    tl->locktype = TRANS_NOLOCK;
	tl->flags = CACHE_CLEAN;
    tl->cache = NULL;
    head = &TID(trans)->translock_hashtab[pagenum & TRANSLOCK_HASHTAB_MASK];
    tl->next = *head;
    *head = tl;

	if (pl->lockid < 0) {
		/* If the page has been top-level cached, the lock has been freed. */
		if ((pl->lockid = NewLock()) < 0)
			return(DBM_TRANS_LOCK_REQUEST_FAILED);
	}
/* XXX A context switch here could cause DbfPageLock pl to be freed (because
   a concurrent releaseLock() sees an nheld of zero... */
	st = (mode == TRANS_READLOCK
		  ? RLOCK(pl->lockid, &grantarg) : WLOCK(pl->lockid, &grantarg));
	if (st < 0) {
		*head = (*head)->next;
		FREE(tl);
		return(DBM_TRANS_LOCK_REQUEST_FAILED);
	}
    tl->locktype = mode;	
	pl->nheld++;

	*tlp = tl;
	return(DBM_OK);
}

static void
freeTransLock PARAMS((tl),
TransLock *tl)
{

	tl->cache = NULL;
	FREE(tl);
}

static DbmRc
addPageLock PARAMS((trans, pagenum, plockp),
DbmTrans *trans				_AND_
PhysPagenum pagenum			_AND_
DbfPageLock **plockp)
{
	DbfPageLock *pl, *plock;
	DbmFile *dbf;

	dbf = DBF(trans);

	if ((plock = (DbfPageLock *)
		 NAMEDMALLOC("DbfPageLock", sizeof(DbfPageLock))) == NULL)
		return(DBM_MALLOC_FAILED);

	plock->pagenum = pagenum;
	plock->cache = NULL;
	plock->nheld = 0;
	plock->read_retained = 0;
	plock->write_retained = 0;

	if ((plock->lockid = NewLock()) < 0) {
		FREE(plock);
		return(DBM_TRANS_LOCK_REQUEST_FAILED);
	}

	/* Put the new lock at the head of the list. */
	pl = dbf->lock_hashtab[pagenum & LOCK_HASHTAB_MASK];
	plock->prev = NULL;
	if (plock->next = pl)		/* Supposed to be an assignment. */
		pl->prev = plock;
	dbf->lock_hashtab[pagenum & LOCK_HASHTAB_MASK] = plock;

	*plockp = plock;
	return(DBM_OK);
}

/*
 * There are no more references to the given page.
 */
static void
removePageLock PARAMS((dbf, pl),
DbmFile *dbf			_AND_
DbfPageLock *pl)
{

	if (pl->next)
		(pl->next)->prev = pl->prev;
	if (pl->prev)
		(pl->prev)->next = pl->next;
	else
		dbf->lock_hashtab[pl->pagenum & LOCK_HASHTAB_MASK] = pl->next;
	if (pl->cache) {
		CacheFree(pl->cache);
		pl->cache = NULL;
	}

	FREE(pl);
}

/*
 * Transaction ARG->TID has requested a MODE lock on ARG->PAGENUM.
 * Consult the locking rules and return 1 if the request can be granted, 0 if
 * the request must block, or -1 if the request must be denied to avoid
 * deadlock.
 */
static int
grant_test PARAMS((arg, mode),
GrantTestArg *arg		_AND_
LOCKMODE mode)
{
	int count, total;
	DbfPageLock *pl;
	PhysPagenum pagenum;
	Tid *req_tid, *tid;
	TransLock *tl;

	req_tid = tid = arg->tid;
	pagenum = arg->pagenum;
	count = 0;
	pl = NULL;

	/*
	 * Search the transaction tree upwards for the page.
	 *
	 * XXX This is a slight modification of Moss's rules: the search is begun
	 * at the requesting transaction instead of at its superior.  I don't
	 * think that this violates one-copy serializability.
	 */
	do {
		tl = tid->translock_hashtab[pagenum & TRANSLOCK_HASHTAB_MASK];
		while (tl && tl->lockinfo->pagenum != pagenum)
			tl = tl->next;
		if (tl) {
			/* Simple deadlock checks. */
			if (tid == req_tid) {
				/* Check self-deadlock. */
				if (mode == LOCK_WRITEMODE && tl->locktype == TRANS_WRITELOCK)
					return(-1);
				if (mode == LOCK_READMODE && tl->locktype == TRANS_WRITELOCK)
					return(-1);
			}
			else if (tl->locktype == TRANS_WRITELOCK
					 || tl->locktype == TRANS_READLOCK)
				return(-1);

			if (mode == LOCK_READMODE
				&& tl->locktype == TRANS_WRITELOCK_RETAINED)
				count++;
			else if (mode == LOCK_WRITEMODE
					 && (tl->locktype == TRANS_WRITELOCK_RETAINED
						 || tl->locktype == TRANS_READLOCK_RETAINED))
				count++;
			/* Save a pointer */
			pl = tl->lockinfo;
		}
	} while (tid = tid->parent);

#ifdef NOTDEF
	/* Top-level transactions are a special case. */
	tid = req_tid;
	if (tid->parent == NULL) {
		tl = tid->translock_hashtab[pagenum & TRANSLOCK_HASHTAB_MASK];
		while (tl && tl->lockinfo->pagenum != pagenum)
			tl = tl->next;
		if (tl)
			pl = tl->lockinfo;
	}
#endif NOTDEF

	if (pl)
		total = ((mode == LOCK_WRITEMODE)
				 ? pl->write_retained + pl->read_retained
				 : pl->write_retained) - count;
	else
		total = 0;

	return(total ? 0 : 1);
}

static DbmRc
addActiveTrans PARAMS((tid),
Tid *tid)
{
	ActiveTrans *active;
	static int hostid = 0;
	static int start_time;
	static int count;

	if (hostid == 0) {
		char hostname[MAXHOSTNAMELEN + 1];
		struct hostent *hostent;

		if (gethostname(hostname, MAXHOSTNAMELEN + 1) < 0
			|| (hostent = gethostbyname(hostname)) == NULL)
			return(DBM_UNIX_ERROR);
		hostid = *(int *) hostent->h_addr;
		time(&start_time);
		count = 0;
	}

	active = tid->active = (ActiveTrans *)
		NAMEDMALLOC("ActiveTrans", sizeof(ActiveTrans));
	if (active == NULL)
		return(DBM_MALLOC_FAILED);
	active->dbm_id = 0;
	active->dbf = NULL;
	active->tidname.hostid = hostid;
	active->tidname.start_time = start_time;
	active->tidname.count = count++;
	active->precommitted = NOT_PRECOMMITTED;
	active->fd = -1;
	active->pathname = NULL;
	active->rootTid = tid;

	if (active_transactions)
		active_transactions->prev = active;
	active->next = active_transactions;
	active->prev = NULL;
	active_transactions = active;

	return(DBM_OK);
}

static void
removeActiveTrans PARAMS((active),
ActiveTrans *active)
{

	if (active->prev)
		(active->prev)->next = active->next;
	else
		active_transactions = active->next;
	if (active->next)
		(active->next)->prev = active->prev;

	FREE(active);
}

static DbmRc
reopenTransFile PARAMS((tid, pathname),
Tid *tid			_AND_
char *pathname)
{
	int fd;

	errno = 0;
	/*
     * This must be opened read only because of the mode on the precommitted
	 * file.  Unfortunately, this means the file cannot be write locked because
	 * SunOS insists that the file be writable to do that.
	 * I don't think this is really a problem since the main dbm file will
	 * (likely) have been write locked before this point.
     */
	if ((fd = OPEN(pathname, O_RDONLY, 0)) < 0)
		return(DBM_OPEN_FAILED);

	TID_FD(tid) = fd;
	tid->active->pathname = STRDUP(pathname);
	return(DBM_OK);
}

static DbmRc
createTransFile PARAMS((tid, flags),
Tid *tid			_AND_
int flags)
{
    int fd, len;
    char *path;
	DbmFile *dbf;
	struct flock flock;
	static long start_time = 0;
	static int count = 1;

	if (start_time == 0)
		time(&start_time);

	dbf = TID_DBF(tid);

    len = strlen(dbf->pathname) + strlen(TRANS_FILE_SUFFIX)
		+ 1 + TRANS_FILE_DATE_LEN + 1 + 10;
    if ((path = (char *) NAMEDMALLOC("TransFile", len)) == NULL)
        return(DBM_MALLOC_FAILED);
    sprintf(path, "%s%s.%.*d.%d", dbf->pathname, TRANS_FILE_SUFFIX,
			TRANS_FILE_DATE_LEN, start_time, count);
	count++;

	errno = 0;
    if ((fd = OPEN(path, flags, TRANS_FILE_CREATE_MODE)) < 0) {
        FREE(path);
        return(DBM_OPEN_FAILED);
    }
	flock.l_type = F_WRLCK;
	flock.l_whence = 0;
	flock.l_start = 0;
	flock.l_len = 0;
	if (fcntl(fd, F_SETLK, &flock) < 0) {
        FREE(path);
		CLOSE(fd);
        return(DBM_FLOCK_FAILED);
    }

	TID_FD(tid) = fd;
	tid->active->pathname = path;
	return(DBM_OK);
}

static DbmRc
getTransFiles PARAMS((pathname, file_list),
char *pathname			_AND_
DbmRecovery **file_list)
{
    int dirname_len, i, n, pattern_len;
    char *dirname, *p, *pattern, *regex;
	struct direct **name_list;
	DbmRecovery *file, *prev;

	if ((p = (char *) rindex(pathname, '/')) == NULL)
		return(DBM_RECOVER_FAILED);
	p++;
	dirname_len = (int) (p - pathname) + 1;
	dirname = (char *) NAMEDMALLOC("TransFile", dirname_len);
	strncpy(dirname, pathname, p - pathname + 1);
	dirname[dirname_len - 1] = '\0';

	pattern_len = strlen(pathname) - (p - pathname) + strlen(TRANS_FILE_SUFFIX)
		+ strlen(TRANS_FILE_RE_EXP) + 1;
	pattern = (char *) NAMEDMALLOC("TransFile", pattern_len);
	sprintf(pattern, "%s%s%s", p, TRANS_FILE_SUFFIX, TRANS_FILE_RE_EXP);
	if (regex = (char *) re_comp(pattern)) {
		fprintf(stderr, "getTransFiles: re_comp(): %s\n", regex);
		FREE(pattern);
		FREE(dirname);
		return(DBM_RECOVER_FAILED);
	}

	n = scandir(dirname, &name_list, matcher, NULL);

	*file_list = NULL;
	prev = NULL;
	for (i = 0; i < n; i++) {
		int len;

		file = (DbmRecovery *) NAMEDMALLOC("TransFile", sizeof(DbmRecovery));
		len = dirname_len + name_list[i]->d_namlen + 1;
		file->pathname = (char *) NAMEDMALLOC("TransFile", len);
		sprintf(file->pathname, "%s%s", dirname, name_list[i]->d_name);
		file->next = NULL;

		if (prev)
			prev->next = file;
		else
			*file_list = file;
		prev = file;
		free(name_list[i]);
	}
	if (n)
		free(name_list);

	FREE(pattern);
	FREE(dirname);
    return(DBM_OK);
}

static int
matcher PARAMS((d),
struct direct *d)
{

	return(re_exec(d->d_name));
}

static int
filesort PARAMS((pptr, qptr),
struct direct **pptr	_AND_
struct direct **qptr)
{
	int plen, qlen, st;
	long pcount, qcount, pnum, qnum;
	char *pp, *pq, *pstart, *qstart;
	struct direct *p, *q;

	/* Names: /a/b/c/prefix.number */
	p = *pptr;
	q = *qptr;
	pstart = (char *) rindex(p->d_name, '.');
	qstart = (char *) rindex(q->d_name, '.');
	pcount = strtol(pstart + 1, &pp, 10);
	qcount = strtol(qstart + 1, &pq, 10);
	while (*(pstart - 1) != '.')
		pstart--;
	while (*(qstart - 1) != '.')
		qstart--;
	pnum = strtol(pstart, &pp, 10);
	qnum = strtol(qstart, &pq, 10);

	plen = pstart - p->d_name;
	qlen = qstart - q->d_name;
	if (plen > qlen)
		return(1);
	if (plen < qlen)
		return(-1);
	if (st = strncmp(p->d_name, q->d_name, plen))
		return(st);

	/* Prefixes are the same. */
	if (pnum > qnum)
		return(1);
	else if (pnum < qnum)
		return(-1);
	if (pcount > qcount)
		return(1);
	else if (pcount < qcount)
		return(-1);
	return(0);		/* Not possible. */
}

static DbmRc
removeTransFile PARAMS((tid),
Tid *tid)
{
	DbmRc rc;

	rc = DBM_OK;
	if (unlink(tid->active->pathname) < 0)
		rc = DBM_UNLINK_FAILED;
	FREE(tid->active->pathname);
	tid->active->pathname = NULL;

	if (CLOSE(TID_FD(tid)) < 0 && rc == DBM_OK)
		rc = DBM_CLOSE_FAILED;

	TID_FD(tid) = -1;
	return(rc);
}

static int tidCount = 0;

static DbmRc
allocTid PARAMS((parent, newtid),
Tid *parent			_AND_
Tid **newtid)
{
	int i;
	DbmRc rc;
	Tid *sibling, *tid;

	if ((tid = (Tid *) NAMEDMALLOC("Tid", sizeof(Tid))) == NULL)
		return(DBM_MALLOC_FAILED);

	if (parent == NULL) {
		if ((rc = addActiveTrans(tid)) != DBM_OK) {
			FREE(tid);
			return(rc);
		}
	}
	else
		tid->active = parent->active;

	tid->state = TRANS_ACTIVE;
	for (i = 0; i < TRANSLOCK_HASHTAB_SIZE; i++)
		tid->translock_hashtab[i] = NULL;
	tid->children = NULL;
	tid->prev = NULL;
	tid->next = NULL;

	/* XXX If parent exists, add this new child to its list */
	if (tid->parent = parent) {
		if (sibling = (tid->parent)->children) {
			tid->next = sibling;
			sibling->prev = tid;
		}
		(tid->parent)->children = tid;
	}

	*newtid = tid;
	tidCount++;
	return(DBM_OK);
}

static void
freeTid PARAMS((tid),
Tid *tid)
{

	if (tid->parent) {
		if (tid->next)
			(tid->next)->prev = tid->prev;
		if (tid->prev)
			(tid->prev)->next = tid->next;
		else
			(tid->parent)->children = tid->next;
	}
	else
		removeActiveTrans(tid->active);

	FREE(tid);
	tidCount--;
}

void
PrintActiveTrans PARAMS((tid),
Tid *tid)
{
	ActiveTrans *active;

	if (tid) {
		active = tid->active;
		fprintf(stderr, "id=%d, path=%s\n",
				active->dbm_id, active->pathname);
	}
	else {
		for (active = active_transactions; active; active = active->next) {
			fprintf(stderr, "id=%d, path=%s, precommitted=%d\n",
					active->dbm_id, active->pathname, active->precommitted);
		}
	}
}

void
ShowDbfLocks NOPARAMS()
{
	int i;
	ActiveTrans *active;
	DbfPageLock *pl;
	DbmFile *dbf;

	for (active = active_transactions; active; active = active->next) {
		if ((dbf = active->dbf) == NULL)
			continue;
		if (active->pathname)
			printf("%s:\n", active->pathname);
		else
			printf("Volatile:\n");
		for (i = 0; i < LOCK_HASHTAB_SIZE; i++) {
			pl = dbf->lock_hashtab[i];
			while (pl) {
				showDbfPageLock(pl);
				pl = pl->next;
			}
		}
	}
}

static void
showDbfPageLock PARAMS((pl),
DbfPageLock *pl)
{

	fprintf(stderr, "pnum=%5d, lockid=%2d, nheld=%2d, retained=%2dr/%2dw",
			pl->pagenum, pl->lockid,
			pl->nheld, pl->read_retained, pl->write_retained);
	if (pl->cache)
		fprintf(stderr, " (nbytes=%7d, flags=%d)",
				pl->cache->nbytes, pl->cache->flags);
	fprintf(stderr, "\n");
	if (pl->cache && (pl->pagenum != pl->cache->pagenum))
		fprintf(stderr, "*** CACHE ERROR: cache->pagenum=%d\n",
				pl->cache->pagenum);
}

void
ShowTransLocks PARAMS((tid),
Tid *tid)
{
	int i, len, max, min, sum;
	PageCache *pc;
	TransLock *tl;

	max = -1;
	min = 1000000;
	sum = 0;
	for (i = 0; i < TRANSLOCK_HASHTAB_SIZE; i++) {
		tl = tid->translock_hashtab[i];
		len = 0;
		while (tl) {
			fprintf(stderr, "lock=%s ",
					tl->locktype == TRANS_NOLOCK ? "None "
					: (tl->locktype == TRANS_READLOCK
					   ? "Read " : "Write"));
			fprintf(stderr, "%s%s",
					tl->flags & CACHE_SHARED ? "(s)" : "   ",
					tl->flags & CACHE_DELETED ? "(d)" : "   ");
			pc = tl->cache;
			if (pc)
				fprintf(stderr, " (pnum=%5d nbytes=%7d flags=%d)",
						pc->pagenum, pc->nbytes, pc->flags);
			fprintf(stderr, "\n");
			len++;
			tl = tl->next;
		}
		if (len < min)
			min = len;
		if (len > max)
			max = len;
		sum += len;
	}
	fprintf(stderr, "Hashtab: min=%d, max=%d, mean=%.1f\n",
				min, max, (double) sum / (double) TRANSLOCK_HASHTAB_SIZE);
}
